from ariths_gen.multi_bit_circuits.subtractors.ripple_carry_subtractor import (
    UnsignedRippleCarrySubtractor, SignedRippleCarrySubtractor
)


from ariths_gen.multi_bit_circuits.subtractors.ripple_borrow_subtractor import (
    UnsignedRippleBorrowSubtractor, SignedRippleBorrowSubtractor
)






from io import StringIO
from ..wire_components import (

    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ..core.arithmetic_circuits import GeneralCircuit
from ..core.cgp_circuit import UnsignedCGPCircuit

from ..one_bit_circuits.logic_gates import (
    AndGate,
    NandGate,
    OrGate,
    NorGate,
    XorGate,
    XnorGate,
    NotGate
)
import re
import random

class ShuffleCircuit(UnsignedCGPCircuit):

    @staticmethod
    def from_circuit(circuit: GeneralCircuit, strategy = "random"):
        o = StringIO()
        circuit.get_cgp_code_flat(o)

        cgp = o.getvalue()
        return ShuffleCircuit(code=cgp.strip(), strategy=strategy, input_widths=[i.N for i in circuit.inputs])
    

    def __init__(self, code: str = "", strategy = "random", input_widths: list = None, inputs: list = None, prefix: str = "", name: str = "cgp", **kwargs):
        """
        Initializes the CGP (Cartesian Genetic Programming) circuit by parsing the provided code and setting up the inputs, 
        internal nodes, and outputs. The circuit is shuffled and maintained as an acyclic forward graph.
        Args:
            code (str): The CGP code string defining the circuit structure.
            input_widths (list): List of widths for each input bus.
            strategy: random, min or max
            inputs (list): List of input Bus objects.
            prefix (str): Prefix for naming components.
            name (str): Name of the circuit.
            **kwargs: Additional keyword arguments for the GeneralCircuit initialization.
        Raises:
            AssertionError: If neither inputs nor input_widths are provided, or if both are provided.
            ValueError: If there are backward or loop connections in the CGP genes, or if output connections are out of range.
        """

        cgp_prefix, cgp_core, cgp_outputs = re.match(
            r"{(.*)}(.*)\(([^()]+)\)", code).groups()

        c_in, c_out, c_rows, c_cols, c_ni, c_no, c_lback = map(
            int, cgp_prefix.split(","))
        
        assert inputs is not None or input_widths is not None, "Either inputs or input_widths must be provided"

        if inputs:
            assert input_widths is None, "Only one of inputs or input_widths must be provided"

            input_widths =[i.N for i in inputs]
            assert sum(input_widths) == c_in, f"CGP input width {c_in} doesn't match inputs {inputs_widths}"



        else:

            assert sum(
                input_widths) == c_in, f"CGP input width {c_in} doesn't match input_widths {input_widths}"
            
            def get_name(i):
                if i < 26:
                    return chr(i + 0x61)
                assert i < 26 * 26
                return chr(i // 26 + 0x60) + chr(i % 26 + 0x61)

            inputs = [Bus(N=bw, prefix=f"input_{get_name(i)}")
                    for i, bw in enumerate(input_widths)]

            # Assign each Bus object in self.inputs to a named attribute of self
            for bus in inputs:
                # Here, bus.prefix is 'input_a', 'input_b', etc.
                # We strip 'input_' and use the remaining part (e.g., 'a', 'b') to create the attribute name
                attr_name = bus.prefix.replace('input_', '')
                setattr(self, attr_name, bus)

        # Adding values to the list
        self.vals = {}
        j = 2  # Start from two, 0=False, 1=True
        for iid, bw in enumerate(input_widths):
            for i in range(bw):
                assert j not in self.vals
                self.vals[j] = inputs[iid].get_wire(i)
                j += 1

        GeneralCircuit.__init__(self, prefix=prefix, name=name, out_N=c_out, inputs=inputs, **kwargs)

        if not code:
            return # only for getting the name
        
        cgp_core = cgp_core.split(")(")



        i = 0

        values = []
        for definition in cgp_core:
            i, in_a, in_b, fn = map(int, re.match(
                r"\(?\[(\d+)\](\d+),(\d+),(\d+)\)?", definition).groups())

            if in_a > i or in_b > i:
                raise ValueError(f"Backward connection in CGP gene \"{definition}\", maxid = {i}")

            if in_a == i or in_b == i:
                raise ValueError(f"Loop connection in CGP gene: \"{definition}\", maxid = {i}")
            
            values.append((i, in_a, in_b, fn))

        connected = []
        disconnected = values.copy()

        to_connect = []

        while disconnected or to_connect:
            
            # get the values that can be connected
            disconnected2 = []
            for i, in_a, in_b, fn in disconnected:
                # pick the nodes that can be connected
                if fn in [8, 9] or (in_a in self.vals and fn in [0, 1]) or (in_a in self.vals and in_b in self.vals):
                    to_connect.append((i, in_a, in_b, fn))
                else:
                    disconnected2.append((i, in_a, in_b, fn))
            
            disconnected = disconnected2

            # print("To connect: ", to_connect)
            # print("Disconnected: ", disconnected)
            # print("Vals", self.vals.keys())
            # print("")


            if strategy == "random":
                # randomly pick one value from to_connect
                picked = random.choice(to_connect)
            elif strategy == "max":
                # select the value with the highest i
                picked = max(to_connect, key=lambda x: x[0])
            elif strategy == "min":
                # select the value with the lowest i
                picked = min(to_connect, key=lambda x: x[0])
            else:
                raise ValueError(f"Unknown strategy {strategy}, can be random, min or max")

            i, in_a, in_b, fn = picked

            to_connect.remove(picked)

            # remove the picked value from disconnected
            comp_set = dict(prefix=f"{self.prefix}_core_{i:03d}", parent_component=self)

            a = b = None

            if fn not in [8, 9]:
                a = self._get_wire(in_a)

                if fn not in [0, 1]:
                    b = self._get_wire(in_b)

            if fn == 0:  # IDENTITY
                o = a
            elif fn == 1:  # NOT
                o = self.add_component(NotGate(a, **comp_set)).out
            elif fn == 2:  # AND
                o = self.add_component(AndGate(a, b, **comp_set)).out
            elif fn == 3:  # OR
                o = self.add_component(OrGate(a, b, **comp_set)).out
            elif fn == 4:  # XOR
                o = self.add_component(XorGate(a, b, **comp_set)).out
            elif fn == 5:  # NAND
                o = self.add_component(NandGate(a, b, **comp_set)).out
            elif fn == 6:  # NOR
                o = self.add_component(NorGate(a, b, **comp_set)).out
            elif fn == 7:  # XNOR
                o = self.add_component(XnorGate(a, b, **comp_set)).out
            elif fn == 8:  # TRUE
                o = ConstantWireValue1()
            elif fn == 9:  # FALSE
                o = ConstantWireValue0()

            assert i not in self.vals
            self.vals[i] = o

        # print("Disconnected: ", disconnected)
        # print("vals: ", self.vals.keys())
        # Output connection
        for i, o in enumerate(map(int, cgp_outputs.split(","))):
            if o >= c_in + c_rows * c_cols + 2:
                raise ValueError(
                    f"Output {i} is connected to wire {o} which is not in the range of CGP wires ({c_in + c_rows * c_cols + 2})")
            w = self._get_wire(o)
            self.out.connect(i, w)




import os
import sys
# Add the parent directory to the system path
DIR_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(DIR_PATH, '..'))

from io import StringIO
from ariths_gen.core.cgp_circuit import UnsignedCGPCircuit
from ariths_gen.wire_components import (
    Wire,
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)

from ariths_gen.core.arithmetic_circuits import GeneralCircuit

from ariths_gen.multi_bit_circuits.others import (
    UnsignedPopCount
)
import numpy as np


def test_popcount(verbose: bool = False):
    """ Test unsigned adders """
    N = 7

    for N in [3, 7, 8, 9, 16]:
        a = Bus(N=N, prefix="a")
        av = np.arange(2**N)

        popcnt = UnsignedPopCount(a=a)
        #o = StringIO()
        #popcnt.get_v_code_hier(o)
        #print(o.getvalue())

        if verbose:
            print(popcnt(av))

        # conv to binary
        r = []
        a_s = av.copy()
        for i in range(N):
            r.append(a_s % 2)
            a_s = a_s // 2
        r = np.dstack(r).reshape(-1, N)

        if verbose:
            print("r = ", r)

        expected = np.sum(r, axis=1)

        if verbose:
            print("expected = ", expected)

        np.testing.assert_array_equal(popcnt(av), expected)

def test_popcount_cgp(verbose: bool = False):
    """ Test unsigned adders """
    N = 7

    for N in [3, 7, 8, 9, 16]:
        a = Bus(N=N, prefix="a")
        av = np.arange(2**N)

        popcnt = UnsignedPopCount(a=a)
        o = StringIO()
        popcnt.get_cgp_code_flat(o)
        cgp = UnsignedCGPCircuit(o.getvalue(), [N])
        v = cgp(av)

        if verbose:
            print(popcnt(av))

        # conv to binary
        r = []
        a_s = av.copy()
        for i in range(N):
            r.append(a_s % 2)
            a_s = a_s // 2
        r = np.dstack(r).reshape(-1, N)

        if verbose:
            print("r = ", r)

        expected = np.sum(r, axis=1)

        if verbose:
            print("expected = ", expected)

        np.testing.assert_array_equal(v, expected)


if __name__ == "__main__":
    test_popcount()
    test_popcount_cgp()
    print("Python popcount tests were successful!")



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit,
)
from ariths_gen.core.logic_gate_circuits import (
    TwoInputLogicGate
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate
)
from ariths_gen.multi_bit_circuits.adders import (
    UnsignedCarryLookaheadAdder
)


class UnsignedBrokenCarrySaveMultiplier(MultiplierCircuit):
    """Class representing unsigned broken carry save (Braun) multiplier.

    It represents an approximative version of unsigned carry save multiplier with simpler structure.
    It is created by modifying an ordinary N-bit unsigned carry save multiplier by omitting partial product
    stage cells by the specified horizontal and vertical cut levels.

    The design promises better area and power parameters in exchange for the loss of computation precision.
    The BCSM design allows to save more partial product stage adders than truncated multiplier architectures.
    ```
                                                    VERTICAL CUT=3

                                                     │
                                            A3B0       A2B0       A1B0       A0B0
                                           ┌───┐     │┌───┐      ┌───┐      ┌───┐
                                           │AND│      │AND│      │AND│      │AND│
                                           └───┘     │└───┘      └───┘      └───┘

                                                     │
                                      A3B1       A2B1       A1B1       A0B1
                                     ┌───┐      ┌───┐│      ┌───┐      ┌───┐
                                     │AND│      │AND│       │AND│      │AND│
                                     └───┘      └───┘│      └───┘      └───┘

                                             ┌────┐  │  ┌────┐     ┌────┐
                                             │    │     │    │     │    │
                                             │ HA │  │  │ HA │     │ HA │
                                             │    │     │    │     │    │
                                             └────┘  │  └────┘     └────┘

              ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─┼─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ HORIZONTAL CUT=2
                      A3B2       A2B2       A1B2          A0B2
                     ┌▼─▼┐      ┌▼─▼┐      ┌▼─▼┐     │   ┌───┐
                     │AND│      │AND│      │AND│         │AND│
                     └┬──┘      └┬──┘      └┬──┘     │   └───┘
                      │          │          │
              A3B3    │  A2B3    │  A1B3    │  A0B3  │  ┌────┐
             ┌▼─▼┐    │ ┌▼─▼┐    │ ┌▼─▼┐    │ ┌▼─▼┐     │    │
             │AND│    │ │AND│    │ │AND│    │ │AND│  │  │ HA │
             └┬──┘    │ └┬──┘    │ └┬──┘    │ └┬──┘     │    │
              │       │  │       │  │       │  │     │  └────┘
              │       │  │       │  │       │  │
              │       │  │       │  │       │  │     │
              │      ┌▼──▼┐     ┌▼──▼┐     ┌▼──▼┐
              │      │    │     │    │     │    │    │
              │    ┌─┤ HA │   ┌─┤ HA │   ┌─┤ HA │
              │    │ │    │   │ │    │   │ │    │    │
       0  0   │    │ └─┬──┘   │ └─┬──┘   │ └──┬─┘
       │  │   │S2  │C2 │S1    │C1 │S0    │C0  │      │
    ┌──▼──▼───▼────▼───▼──────▼───▼──────▼──┐ │
    │            Carry-propagate            │ │      │
    │                adder                  │ │
    └─┬───────┬──────────┬──────────┬───────┘ │      │
      │       │          │          │         │
      │       │          │          │         │      │
      │       │          │          │         │
      │       │          │          │         │      │
      ▼       ▼          ▼          ▼         ▼            ▼         ▼        ▼
      P7      P6         P5         P4        P3     │     P2=0      P1=0     P0=0
    ```
    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        horizontal_cut (int, optional): Specifies horizontal cut used in broken carry save multiplier circuit creation. Defaults to 0.
        vertical_cut (int, optional): Specifies vertical cut used in broken carry save multiplier circuit creation. Defaults to 0.
        prefix (str, optional): Prefix name of unsigned broken carry save multiplier. Defaults to "".
        name (str, optional): Name of unsigned broken carry save multiplier. Defaults to "u_bamcsa".
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used for final vector merging of sums and carries. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, horizontal_cut: int = 0, vertical_cut: int = 0, prefix: str = "", name: str = "u_bamcsa", unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        # NOTE: If horizontal/vertical cut is specified as 0 the final circuit is a simple carry save multiplier
        self.horizontal_cut = horizontal_cut
        self.vertical_cut = vertical_cut

        self.N = max(a.N, b.N)
        # Horizontal cut level should be: 0 <= horizontal_cut < N
        # Vertical cut level should be: horizontal_cut <= vertical_cut < 2*N
        assert horizontal_cut < self.N
        assert vertical_cut < 2*self.N

        # Vertical cut should be greater or equal to horizontal cut
        assert vertical_cut >= horizontal_cut

        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        self.ommited_rows = self.horizontal_cut
        # Gradual generation of partial products
        for b_multiplier_index in range(self.horizontal_cut, self.N):
            # Number of elements that should be ommited in the current level based on vertical cut
            pp_row_elems_to_skip = self.vertical_cut - b_multiplier_index if self.vertical_cut - b_multiplier_index > 0 else 0
            # Number of pp pairs present in the current row
            pp_row_elems = self.N-pp_row_elems_to_skip if self.N-pp_row_elems_to_skip > 0 else 0
            self.ommited_rows += 1 if pp_row_elems == 0 else 0

            row_pp_ord = 0
            for a_multiplicand_index in range((self.N-pp_row_elems), self.N):
                row_pp_ord += 1
                # AND gates generation for calculation of partial products
                obj_and = AndGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_and"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                self.add_component(obj_and)

                if a_multiplicand_index == 0:
                    self.out.connect(a_multiplicand_index + b_multiplier_index, obj_and.out)

                # HAs generation for first row of adders (the second row in total, first contains just pp generation)
                if b_multiplier_index == self.horizontal_cut+1 and a_multiplicand_index != self.N-1:
                    previous_product = self.get_previous_component(self.N-(pp_row_elems_to_skip+1)+row_pp_ord).out
                    obj_adder = HalfAdder(a=obj_and.out, b=previous_product, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_adder)

                    if (a_multiplicand_index == 0):
                        self.out.connect(b_multiplier_index, obj_adder.get_sum_wire())

                # FAs generation for the remaining rows
                elif b_multiplier_index > self.horizontal_cut+1 and a_multiplicand_index != self.N-1:
                    previous_sum = self.get_previous_component((self.N-(pp_row_elems_to_skip+1))*2-1).get_sum_wire() if a_multiplicand_index != self.N-2 else self.get_previous_component((self.N-(pp_row_elems_to_skip+1))*2).out

                    if a_multiplicand_index == self.N-pp_row_elems and b_multiplier_index <= vertical_cut:
                        obj_adder = HalfAdder(a=obj_and.out, b=previous_sum, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))

                        self.out.connect(b_multiplier_index, obj_adder.get_sum_wire())
                    else:
                        previous_carry = self.get_previous_component((row_pp_ord*2-1)+((self.N-(a_multiplicand_index+1))*2)).get_carry_wire()
                        obj_adder = FullAdder(a=obj_and.out, b=previous_sum, c=previous_carry, prefix=self.prefix+"_fa"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))

                    self.add_component(obj_adder)
                    if (a_multiplicand_index == 0):
                        self.out.connect(b_multiplier_index, obj_adder.get_sum_wire())

                # PRODUCT GENERATION
                # 1 bit multiplier case
                if (a_multiplicand_index == 0 and b_multiplier_index == self.horizontal_cut) and (self.horizontal_cut + self.ommited_rows == self.N-1):
                    self.out.connect(a_multiplicand_index + b_multiplier_index, obj_and.out)

                    if a_multiplicand_index == self.N-1 and b_multiplier_index == self.N-1:
                        self.out.connect(a_multiplicand_index+b_multiplier_index+1, ConstantWireValue0())

                # If architecture is cut to just a row of AND gates
                elif self.ommited_rows == self.N-1:
                    self.out.connect(b_multiplier_index + a_multiplicand_index, obj_and.out)

                    if a_multiplicand_index == self.N-1:
                        self.out.connect(self.out.N-1, ConstantWireValue0())

        # Connecting the output bits generated from ommited cells to ground
        if self.vertical_cut == 2*self.N-1:
            [self.out.connect(out_id, ConstantWireValue0()) for out_id in range(self.out.N)]
        else:
            for grounded_out_index in range(0, max(self.horizontal_cut, self.vertical_cut)):
                self.out.connect(grounded_out_index, ConstantWireValue0())

        # Final addition of remaining bits using chosen unsigned multi bit adder
        final_cpa_N = self.out.N - max(self.N, self.vertical_cut+1)
        if self.ommited_rows != self.N-1:
            previous_sums = []
            previous_carries = []
            for wire_id in range(final_cpa_N):
                if (wire_id == final_cpa_N-1):
                    previous_sums.append(ConstantWireValue0())
                else:
                    prev_sum_obj = self.get_previous_component((final_cpa_N-2-wire_id)*2) if wire_id < final_cpa_N-2 else self.get_previous_component()
                    if isinstance(prev_sum_obj, TwoInputLogicGate):
                        previous_sums.append(prev_sum_obj.out)
                    else:
                        previous_sums.append(prev_sum_obj.get_sum_wire())
                if (wire_id == final_cpa_N-1 or (self.vertical_cut+1 == self.out.N-final_cpa_N and wire_id == 0)):
                    previous_carries.append(ConstantWireValue0())
                else:
                    prev_carry_obj = self.get_previous_component((final_cpa_N-2-wire_id)*2+2)
                    if isinstance(prev_carry_obj, TwoInputLogicGate):
                        previous_carries.append(prev_carry_obj.out)
                    else:
                        previous_carries.append(prev_carry_obj.get_carry_wire())

            # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
            adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(final_cpa_N)
            adder_a = Bus(prefix=f"a", wires_list=previous_sums)
            adder_b = Bus(prefix=f"b", wires_list=previous_carries)
            final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True)
            self.add_component(final_adder)
            [self.out.connect(o, final_adder.out.get_wire(o-max(self.N, self.vertical_cut)), inserted_wire_desired_index=o-max(self.N, self.vertical_cut)) for o in range(max(self.N, self.vertical_cut), self.out.N)]



import os
import sys
# Add the parent directory to the system path
DIR_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(DIR_PATH, '..'))

from ariths_gen.wire_components import (
    Wire,
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)

from ariths_gen.core.arithmetic_circuits import GeneralCircuit

from ariths_gen.multi_bit_circuits.others import (
    PopCountCompare
)

from ariths_gen.core.cgp_circuit import UnsignedCGPCircuit, SignedCGPCircuit


import numpy as np
import math
from io import StringIO

from itertools import product

def test_popcountcompare_same(verbose: bool = False):
    N = 10

    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")

    test_cases = list(product(list(range(2**N)), repeat=2))
    all = np.array(test_cases)

    if verbose:
        print(all)

    av = all[:, 0]
    bv = all[:, 1]
    def popcnt(x):
        mask = x & (2**np.arange(N)).reshape(-1, 1)
        return np.sum(mask > 0, axis=0)
    cnta = (popcnt(av))
    cntb = (popcnt(bv))

    if verbose:
        print(cnta)
        print(cntb)

    cmp = PopCountCompare(a=a, b=b)
    cmp.get_v_code_hier(open("tmp.verilog", "w"))

    v = cmp(av, bv)

    if verbose:
        print("ret = ", v)

    expected = np.array(cnta >= cntb).astype(int)

    if verbose:
        print("expected = ", expected)

    #expected = np.sum(r, axis=1)
    np.testing.assert_array_equal(v, expected)


def test_popcountcompare_small(verbose: bool = False):
    N = 10

    a = Bus(N=N, prefix="a")
    b = Bus(N=N // 2, prefix="b")

    test_cases = list(product(range(2**N), range(2**(N//2))))
    all = np.array(test_cases)

    if verbose:
        print(all)

    av = all[:, 0]
    bv = all[:, 1]
    def popcnt(x):
        mask = x & (2**np.arange(N)).reshape(-1, 1)
        return np.sum(mask > 0, axis=0)
    cnta = (popcnt(av))
    cntb = (popcnt(bv))

    if verbose:
        print(cnta)
        print(cntb)

    cmp = PopCountCompare(a=a, b=b)
    cmp.get_v_code_hier(open("tmp.verilog", "w"))

    v = cmp(av, bv)

    if verbose:
        print("ret = ", v)

    expected = np.array(cnta >= cntb).astype(int)

    if verbose:
        print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)
    np.testing.assert_array_equal(v, expected)


def test_popcountcompare_small2(verbose: bool = False):
    N = 10

    a = Bus(N=N // 2, prefix="a")
    b = Bus(N=N, prefix="b")

    test_cases = list(product( range(2**(N//2)), range(2**N)))
    all = np.array(test_cases)

    if verbose:
        print(all)

    av = all[:, 0]
    bv = all[:, 1]
    def popcnt(x):
        mask = x & (2**np.arange(N)).reshape(-1, 1)
        return np.sum(mask > 0, axis=0)
    cnta = (popcnt(av))
    cntb = (popcnt(bv))

    if verbose:
        print(cnta)
        print(cntb)

    cmp = PopCountCompare(a=a, b=b)
    cmp.get_v_code_hier(open("tmp.verilog", "w"))

    v = cmp(av, bv)

    if verbose:
        print("ret = ", v)

    expected = np.array(cnta >= cntb).astype(int)

    if verbose:
        print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)
    np.testing.assert_array_equal(v, expected)


def test_popcountcompare_small2_cgp(verbose: bool = False):
    N = 10

    a = Bus(N=N // 2, prefix="a")
    b = Bus(N=N, prefix="b")

    test_cases = list(product( range(2**(N//2)), range(2**N)))
    all = np.array(test_cases)

    if verbose:
        print(all)

    av = all[:, 0]
    bv = all[:, 1]
    def popcnt(x):
        mask = x & (2**np.arange(N)).reshape(-1, 1)
        return np.sum(mask > 0, axis=0)
    cnta = (popcnt(av))
    cntb = (popcnt(bv))

    if verbose:
        print(cnta)
        print(cntb)

    cmp = PopCountCompare(a=a, b=b)
    cmp.get_v_code_hier(open("tmp.verilog", "w"))
    cmp.get_cgp_code_flat(open("tmp.cgp", "w"))

    o = StringIO()
    cmp.get_cgp_code_flat(o)
    cgp = UnsignedCGPCircuit(o.getvalue(), [N//2, N])

    v = cgp(av, bv)

    if verbose:
        print("ret = ", v)

    expected = np.array(cnta >= cntb).astype(int)

    if verbose:
        print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)
    np.testing.assert_array_equal(v, expected)


if __name__ == "__main__":
    test_popcountcompare_same()
    test_popcountcompare_small()
    test_popcountcompare_small2()
    test_popcountcompare_small2_cgp()
    print("Python popcount_compare tests were successful!")




from ariths_gen.multi_bit_circuits.multipliers import UnsignedArrayMultiplier
import numpy as np

from ariths_gen.tools.shuffle_circuit import ShuffleCircuit
from ariths_gen.wire_components import Bus
from io import StringIO


def test_shuffle_circuit():
    a = Bus(N=4, prefix="a")
    b = Bus(N=4, prefix="b")
    m = UnsignedArrayMultiplier(a, b, prefix="m")

    na = np.arange(0, 2**4).reshape(-1, 1)
    nb = np.arange(0, 2**4).reshape(1, -1)

    assert(np.all(m(na, nb) == na * nb))

    o = StringIO()
    m.get_cgp_code_flat(o)

    shuffled = ShuffleCircuit.from_circuit(m, strategy="random")
    assert(np.all(shuffled(na, nb) == na * nb))
    

def test_shuffle_cgp():
    a = Bus(N=4, prefix="a")
    b = Bus(N=4, prefix="b")
    m = UnsignedArrayMultiplier(a, b, prefix="m")

    na = np.arange(0, 2**4).reshape(-1, 1)
    nb = np.arange(0, 2**4).reshape(1, -1)

    assert(np.all(m(na, nb) == na * nb))

    o = StringIO()
    m.get_cgp_code_flat(o)

    cgp = o.getvalue()

    shuffled = ShuffleCircuit(code=cgp.strip(), input_widths=[4, 4])
    assert(np.all(shuffled(na, nb) == na * nb))

def test_shuffle_strategies():
    a = Bus(N=4, prefix="a")
    b = Bus(N=4, prefix="b")
    m = UnsignedArrayMultiplier(a, b, prefix="m")

    na = np.arange(0, 2**4).reshape(-1, 1)
    nb = np.arange(0, 2**4).reshape(1, -1)

    assert(np.all(m(na, nb) == na * nb))

    o = StringIO()
    m.get_cgp_code_flat(o)

    cgp = o.getvalue()

    shuffled = ShuffleCircuit(code=cgp.strip(), input_widths=[4, 4], strategy="min")
    assert(np.all(shuffled(na, nb) == na * nb))

    shuffled = ShuffleCircuit(code=cgp.strip(), input_widths=[4, 4], strategy="max")
    assert(np.all(shuffled(na, nb) == na * nb))

    shuffled = ShuffleCircuit(code=cgp.strip(), input_widths=[4, 4], strategy="random")
    assert(np.all(shuffled(na, nb) == na * nb))


if __name__ == "__main__":
    test_shuffle_cgp()
    test_shuffle_circuit()
    test_shuffle_strategies()


from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate
)


class UnsignedTruncatedArrayMultiplier(MultiplierCircuit):
    """Class representing unsigned truncated array multiplier.

    It represents an approximative version of unsigned array multiplier with simpler structure.
    It is created by modifying an ordinary N-bit unsigned array multiplier by ignoring
    (truncating) some of the partial products.

    The design promises better area and power parameters in exchange for the loss of computation precision.
    ```
                                                           CUT=2
                                           A3B0    A2B0      │   A1B0    A0B0
                                          ┌───┐   ┌───┐         ┌───┐   ┌───┐
                                          │AND│   │AND│      │  │AND│   │AND│
                                          └───┘   └───┘         └───┘   └───┘
                                                     ┌ ─ ─ ─ ┘
                                     A3B1       A2B1       A1B1       A0B1
                                    ┌───┐      ┌───┐ │    ┌───┐      ┌───┐
                                    │AND│      │AND│      │AND│      │AND│
                                    └───┘      └───┘ │    └───┘      └───┘
                                 ┌────┐     ┌────┐     ┌────┐     ┌────┐
                                 │    │     │    │   │ │    │     │    │
                                 │ HA │     │ FA │     │ FA │     │ HA │
                                 │    │     │    │   │ │    │     │    │
                                 └────┘     └────┘     └────┘     └────┘
              ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─┬ ─ ─ ─ ─ ┴─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ CUT=2
                       A3B2         A2B2         A1B2       A0B2
                      ┌▼─▼┐        ┌▼─▼┐   │    ┌───┐      ┌───┐
                      │AND│        │AND│        │AND│      │AND│
                      └┬──┘        └┬──┘   │    └───┘      └───┘
                       │            │        ┌────┐     ┌────┐
                       │            │      │ │    │     │    │
                       │        ┌ ─ ┼─ ─ ─ ┘ │ FA │     │ HA │
                       │            │        │    │     │    │
                       │        │   │        └────┘     └────┘
               A3B3    │  A2B3      │  A1B3      A0B3
              ┌▼─▼┐    │ ┌▼─▼┐  │   │ ┌───┐     ┌───┐
              │AND│    │ │AND│      │ │AND│     │AND│
              └┬──┘    │ └┬──┘  │   │ └───┘     └───┘
           ┌───▼┐     ┌▼──▼┐       ┌┼───┐    ┌────┐
           │    │     │    │    │  ││   │    │    │
    ┌──────┤ HA │◄────┤ HA │       ││FA │    │ HA │
    │      │    │     │    │    │  ││   │    │    │
    │      └──┬─┘     └──┬─┘       └┼───┘    └────┘
    │         │          │      │   │
    ▼         ▼          ▼          ▼           ▼          ▼          ▼          ▼
    P7        P6         P5     │   P4          P3=0       P2=0       P1=0       P0=0
    ```
    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        truncation_cut (int, optional): Specifies truncation cut level used in the truncated array multiplier circuit creation. Note: If equal to 0, the final circuit behaves as an ordinary array multiplier. Defaults to 0.
        prefix (str, optional): Prefix name of unsigned truncated array multiplier. Defaults to "".
        name (str, optional): Name of unsigned truncated array multiplier. Defaults to "u_tm".
    """
    def __init__(self, a: Bus, b: Bus, truncation_cut: int = 0, prefix: str = "", name: str = "u_tm", **kwargs):
        # NOTE: If truncation_cut is specified as 0 the final circuit is a simple array multiplier
        self.truncation_cut = truncation_cut

        self.N = max(a.N, b.N)
        # Cut level should be: 0 <= truncation_cut < N
        assert truncation_cut < self.N

        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # Gradual generation of partial products
        for b_multiplier_index in range(self.truncation_cut, self.N):
            for a_multiplicand_index in range(self.truncation_cut, self.N):
                # AND gates generation for calculation of partial products
                obj_and = AndGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_and"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                self.add_component(obj_and)

                if b_multiplier_index != self.truncation_cut:
                    previous_product = self.components[a_multiplicand_index + b_multiplier_index - 2*self.truncation_cut].out if b_multiplier_index == self.truncation_cut + 1 else self.get_previous_partial_product(a_index=a_multiplicand_index, b_index=b_multiplier_index, mult_type="tm")

                    # HA generation for first 1-bit adder in each row starting from the second one
                    if a_multiplicand_index == self.truncation_cut:
                        obj_adder = HalfAdder(self.get_previous_component().out, previous_product, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)
                        # Product generation
                        self.out.connect(b_multiplier_index + self.truncation_cut, obj_adder.get_sum_wire())

                    # HA generation, last 1-bit adder in second row
                    elif a_multiplicand_index == self.N-1 and b_multiplier_index == self.truncation_cut+1:
                        obj_adder = HalfAdder(self.get_previous_component().out, self.get_previous_component(number=2).get_carry_wire(), prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)

                    # FA generation
                    else:
                        obj_adder = FullAdder(self.get_previous_component().out, previous_product, self.get_previous_component(number=2).get_carry_wire(), prefix=self.prefix+"_fa"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)

                # PRODUCT GENERATION
                if (a_multiplicand_index == self.truncation_cut and b_multiplier_index == self.truncation_cut) or (self.truncation_cut == self.N-1):
                    self.out.connect(a_multiplicand_index + b_multiplier_index, obj_and.out)

                    # 1 bit multiplier case
                    if a_multiplicand_index == self.N-1 and b_multiplier_index == self.N-1:
                        self.out.connect(a_multiplicand_index+b_multiplier_index+1, ConstantWireValue0())

                elif b_multiplier_index == self.N-1 and self.truncation_cut != self.N-1:
                    self.out.connect(b_multiplier_index + a_multiplicand_index, obj_adder.get_sum_wire())

                    if a_multiplicand_index == self.N-1:
                        self.out.connect(self.out.N-1, obj_adder.get_carry_wire())

        # Connecting the output bits generated from ommited cells to ground
        for grounded_out_index in range(0, self.truncation_cut*2):
            self.out.connect(grounded_out_index, ConstantWireValue0())



from ariths_gen.wire_components import (
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit,
    GeneralCircuit
)
from ariths_gen.multi_bit_circuits.adders import (
    UnsignedRippleCarryAdder
)
from typing import Optional
from math import log2, ceil


class UnsignedPopCount(GeneralCircuit):
    """Class representing unsigned popcount circuit.

    Popcount circuit is a circuit that counts the number of 1s in a binary number.

    """

    def __init__(self, a: Bus, adder : Optional[GeneralCircuit] = None, prefix : str = "", name : str = "popcnt", **kwargs):
        self.N = a.N
        outc = ceil(log2(self.N + 1))
        #print("outc", outc)
        super().__init__(name=name, prefix=prefix, inputs=[a], out_N=outc)

        self.a.bus_extend(2**(outc - 1), prefix=a.prefix)
        #print(self.a)
        self.adder = adder
        if not self.adder:
            self.adder = UnsignedRippleCarryAdder

        # tree reduction
        def create_tree(a: Bus, depth: int, branch="A"):
            #print(a)
            if a.N == 1:
                return a
            else:
                half = a.N // 2
                b_in = Bus(N=half, prefix=f"b_inn{branch}_{depth}A")
                c_in = Bus(N=a.N - half, prefix=f"c_inn{branch}_{depth}B")
                #print(a, half, a.N)
                for i, j in enumerate(range(half)):
                    b_in.connect(i, a.get_wire(j))
                for i, j in enumerate(range(half, a.N)):
                    c_in.connect(i, a.get_wire(j))

                b = create_tree(b_in, depth=depth + 1, branch = branch + "A")
                c = create_tree(c_in, depth= depth + 1, branch = branch + "B")
                d = self.adder(a=b, b=c, 
                               prefix = f"{self.prefix}_add{branch}_{depth}", inner_component=True,
                               parent_component=self)
                self.add_component(d)
                return d.out

        sumbus = create_tree(self.a,0, "X")
        #print(sumbus)
        self.out.connect_bus(sumbus)



from .wires import Wire, ConstantWireValue0, ConstantWireValue1


class Bus():
    """Class representing bus of wires used as inputs/outputs of bigger circuits.

    Description of the __init__ method.

    Args:
        prefix (str, optional): Prefix name of the bus. Defaults to "bus".
        N (int, optional): Number of wires in the bus. Defaults to 1.
        wires_list (list, optional): List of Wire objects used to clone one bus to another. Defaults to None.
        out_bus (bool, optional): Specifies whether this Bus is an output bus of some previous component. Defaults to False.
        signed (bool, optional): Specifies whether this Bus should consider signed numbers or not (used for C code generation). Defaults to False.
    """
    def __init__(self, prefix: str = "bus", N: int = 1, wires_list: list = None, out_bus: bool = False, signed: bool = False):
        self.out_bus = out_bus
        if wires_list is None:
            self.prefix = prefix
            # Adding wires into current bus's wires list (wire names are concatenated from bus prefix and their index position inside the bus in square brackets)
            self.bus = [Wire(name=prefix+f"[{i}]" if N != 1 else prefix, prefix=prefix, index=i, parent_bus=self) for i in range(N)]
            self.N = N
        else:
            self.prefix = prefix
            self.bus = wires_list
            self.N = len(self.bus)

        # Determine C code signedness
        self.signed = signed
        if self.N > 8:
            self.c_var_size = 64
            if signed is True:
                self.c_type = "int64_t"
            else:
                self.c_type = "uint64_t"
        else:
            self.c_var_size = 8
            if signed is True:
                self.c_type = "int8_t"
            else:
                self.c_type = "uint8_t"

    def is_output_bus(self):
        """Tells whether this Bus is an output bus.

        Returns:
            bool: Returns True if it is an output bus of some component.
        """
        return self.out_bus

    def bus_extend(self, N: int, prefix: str = "bus", desired_extension_wire: Wire = ConstantWireValue0()):
        """Provides bus extension to contain more wires.

        Args:
            N (int): Number of wires in the bus. Defaults to 1.
            prefix (str, optional): Prefix name of the bus. Defaults to "bus".
            desired_extension_wire (Wire, optional): Specifies the wire that should be connected to all of the extending bus wires. Defaults to ConstantWireValue0().
        """
        # Checks if any extension is neccesarry and if so, proceeds to wire extend the bus
        if self.N < N:
            # Adding wires into current bus's wires list (wire names are concatenated from bus prefix and their index position inside the bus in square brackets)
            self.bus += [Wire(name=prefix+f"[{i}]", prefix=prefix, index=i, parent_bus=self) for i in range(self.N, N)]

            for w_index in range(self.N, N):
                self.connect(bus_wire_index=w_index, inner_component_out_wire=desired_extension_wire)

            self.N = N

    def get_wire(self, wire_index: int = 0):
        """Retrieves a wire from the bus by a given index.

        Args:
            wire_index (int, optional): Index of wire to be retrieved from the bus. Defaults to 0.

        Returns:
            Wire: Returning wire from the bus.
        """
        assert wire_index < self.N, f"Wire index {wire_index} is out of bounds of the bus {self.prefix} with size {self.N}"
        return self.bus[wire_index]

    def __getitem__(self, i):
        return self.get_wire(i)

    # Connecting output wire of the inner circuit component to desired position in the described circuit's output bus
    def connect(self, bus_wire_index: int, inner_component_out_wire: Wire, inserted_wire_desired_index: int = -1):
        """Connects given 'Wire' object to a 'bus_wire_index' within this bus.

        Used for connection of output wire of the inner circuit component
        to the appropriate wire of the circuit's output bus.

        Args:
            bus_wire_index (int): Index in bus to store given wire in.
            inner_component_out_wire (Wire): Wire of some other component (mostly its output) to store in the bus.
            inserted_wire_desired_index(int, optional): Optional desired explicit index, where 'inner_component_out_wire' value resides in the inner components's output bus. Otherwise 'inner_component_out_wire' self index value is used. Defaults to -1.
        """
        inserted_wire_index = inserted_wire_desired_index if inserted_wire_desired_index != -1 else inner_component_out_wire.index
        # Used for connection of constant wire value into a bus
        if inner_component_out_wire.is_const():
            self.bus[bus_wire_index] = inner_component_out_wire
        # Proper connection of wires that themselves are not yet a member of any other bus and also those that could be part of some bus but do not have `inserted_wire_desired_index` defined
        elif inner_component_out_wire.parent_bus is None or inserted_wire_desired_index == -1:
            self.bus[bus_wire_index] = Wire(name=inner_component_out_wire.name, prefix=inner_component_out_wire.prefix, index=inserted_wire_index, value=inner_component_out_wire.value, parent_bus=self)
        # Proper connection of wires that are already a member of some other bus and are desired to connect value from their previous bus to this one at desired index position
        elif inserted_wire_desired_index != -1:
            self.bus[bus_wire_index] = Wire(name=inner_component_out_wire.name, prefix=inner_component_out_wire.parent_bus.prefix, index=inserted_wire_index, value=inner_component_out_wire.value, parent_bus=self)

    def __setitem__(self, i, v):
        self.connect(i, v)

    def connect_bus(self, connecting_bus: object, start_connection_pos: int = 0, end_connection_pos: int = -1, offset: int = 0):
        """Ensures connection of specified bus wires to this bus wires.

        Used for connection of some inner circuit component's output bus (`connecting_bus`) wires
        to the appropriate input bus (this `self` bus) wires of some other circuit.

        Args:
            connecting_bus (object): Specifies the connecting bus.
            start_connection_pos (int, optional): Specifies the position from which to start interconnecting wires from the `connecting_bus` to this `self` bus. Defaults to 0.
            end_connection_pos (int, optional): Specifies the position from which to end interconnecting wires from the `connecting_bus` to this `self` bus. Defaults to -1.
            offset (int, optional): Specifies the offset wire index position in the `self` bus for proper connection (i.e. wire at index position 5 in the `connecting_bus` with offset set to 5 will be connected to `self` bus index position 0). Default to 0.
        """
        if end_connection_pos == -1:
            end_connection_pos = self.N

        [self.connect(o-offset, connecting_bus.get_wire(o), inserted_wire_desired_index=o) for o in range(start_connection_pos, end_connection_pos)]

    """ PYTHON CODE GENERATION """
    def return_bus_wires_values_python_flat(self):
        """Retrieves values from bus's wires and stores them in bus's corresponding Python variable (object) at proper offset bit position in the bus for flat generation.

        Returns:
            str: Python code for assigning wire values into bus represented in Python code variable.
        """
        # Ensures correct binding between the bus wire index and the wire itself
        # It is used for the case when multiple of the same wire (e.g. `ContantWireValue0()`) are present in the bus (its id would otherwise be incorrect when using `self.bus.index(_)`)
        mapped_positions = [(w_id, self.bus[w_id]) for w_id in range(self.N)]
        return "".join([f"  {self.prefix} = 0\n"] + [f"  {self.prefix} = ({self.prefix}) | {w[1].return_wire_value_python_flat(offset=w[0])}" for w in mapped_positions])

    def return_bus_wires_sign_extend_python_flat(self, retype: bool = False):
        """Sign extends the bus's corresponding Python variable (object) to ensure proper flat Python code variable signedness.

        Returns:
            str: Python code for sign extending the bus variable wire values.
        """
        if self.signed is True:
            last_bus_wire = self.bus[-1]

            assert self.N < 64, "Sign extension is not supported for bus with more than 64 bits"
            if retype:
                rewrite = f"""
  if hasattr({self.prefix}, 'astype'):
    {self.prefix} = {self.prefix}.astype("int64")
  else:
    from ctypes import c_int64
    {self.prefix} = c_int64({self.prefix}).value\n"""
            else:
                rewrite = ""
            
            return "".join([f"  {self.prefix} = ({self.prefix}) | {last_bus_wire.return_wire_value_python_flat(offset=i)}" for i in range(len(self.bus), 64)]) + rewrite
        else:
            return ""

    def __str__(self):
        return f"<wire N={self.N} prefix={self.prefix} \"" + (",".join([str(w) for w in self.bus])) + "\">"

    """ C CODE GENERATION """
    def get_declaration_c(self):
        """Bus declaration in C code.

        Returns:
            str: C code for declaration and initialization of bus name.
        """
        return f"  {self.c_type} {self.prefix} = 0;\n"

    def return_bus_wires_values_c_flat(self):
        """Retrieves values from bus's wires and stores them in bus's corresponding C variable at proper offset bit position in the bus for flat generation.

        Returns:
            str: C code for assigning wire values into bus represented in C code variable.
        """
        # Ensures correct binding between the bus wire index and the wire itself
        # It is used for the case when multiple of the same wire (e.g. `ContantWireValue0()`) are present in the bus (its id would otherwise be incorrect when using `self.bus.index(_)`)
        mapped_positions = [(w_id, self.bus[w_id]) for w_id in range(self.N)]
        return "".join([f"  {self.prefix} |= {w[1].return_wire_value_c_flat(offset=w[0])}" for w in mapped_positions])

    def return_bus_wires_values_c_hier(self):
        """Retrieves values from bus's wires and stores them in bus's corresponding C variable at proper offset bit position in the bus for hierarchical generation.

        Returns:
            str: C code for assigning wire values into bus represented in C code variable.
        """
        # Ensures correct binding between the bus wire index and the wire itself
        # It is used for the case when multiple of the same wire (e.g. `ContantWireValue0()`) are present in the bus (its id would otherwise be incorrect when using `self.bus.index(_)`)
        mapped_positions = [(w_id, w) for w_id, w in enumerate(self.bus) if ((w.parent_bus is None) or (w.parent_bus is not None and w.prefix != self.prefix) or (w.is_const()))]
        return "".join([f"  {self.prefix} |= {w[1].return_wire_value_c_hier(offset=w[0])}" for w in mapped_positions])

    def return_bus_wires_sign_extend_c_flat(self):
        """Sign extends the bus's corresponding C variable to ensure proper flat C code variable signedness.

        Returns:
            str: C code for sign extending the bus variable wire values.
        """
        if self.signed is True:
            last_bus_wire = self.bus[-1]
            return "".join([f"  {self.prefix} |= {last_bus_wire.return_wire_value_c_flat(offset=i)}" for i in range(len(self.bus), self.c_var_size)])
        else:
            return ""

    def return_bus_wires_sign_extend_c_hier(self):
        """Sign extends the bus's corresponding C variable to ensure proper hier C code variable signedness.

        Returns:
            str: C code for sign extending the bus variable wire values.
        """
        if self.signed is True:
            last_bus_wire = self.bus[-1]
            return "".join([f"  {self.prefix} |= {last_bus_wire.return_wire_value_c_hier(offset=i)}" for i in range(len(self.bus), self.c_var_size)])
        else:
            return ""

    """ VERILOG CODE GENERATION """
    def return_bus_wires_values_v_flat(self):
        """Retrieves values from bus's wires and stores them in bus's corresponding Verilog variable at proper offset bit position in the bus for flat generation.

        Returns:
            str: Verilog code for assigning wire values into bus represented in Verilog code bus variable.
        """
        # Ensures correct binding between the bus wire index and the wire itself
        # It is used for the case when multiple of the same wire (e.g. `ContantWireValue0()`) are present in the bus (its id would otherwise be incorrect when using `self.bus.index(_)`)
        mapped_positions = [(w_id, self.bus[w_id]) for w_id in range(self.N)]
        return "".join([f"  assign {self.prefix}[{w[0]}] = {w[1].return_wire_value_v_flat()}" for w in mapped_positions])

    def return_bus_wires_values_v_hier(self):
        """Retrieves values from bus's wires and stores them in bus's corresponding Verilog variable at proper offset bit position in the bus for hierarchical generation.

        Returns:
            str: Verilog code for assigning wire values into bus represented in Verilog code variable.
        """
        # Ensures correct binding between the bus wire index and the wire itself
        # It is used for the case when multiple of the same wire (e.g. `ContantWireValue0()`) are present in the bus (its id would otherwise be incorrect when using `self.bus.index(_)`)
        mapped_positions = [(w_id, w) for w_id, w in enumerate(self.bus) if ((w.parent_bus is None) or (w.parent_bus is not None and w.prefix != self.prefix) or (w.is_const()))]
        return "".join([f"  assign {self.prefix}[{w[0]}] = {w[1].return_wire_value_v_hier()}" for w in mapped_positions])

    def get_unique_assign_out_wires_v(self, circuit_block: object):
        """Returns bus's wires used for hierarchical one bit subcomponent's function block invocation and output wires assignments.

        Args:
            circuit_block (object): Object describing corresponding function block that is being invoked for proper output wires assignment during instantiation.

        Returns:
            str: Verilog code unique bus wires for proper subcomponent's function block invocation.
        """
        unique_out_wires = []
        [unique_out_wires.append(w.prefix) if w.prefix not in unique_out_wires else None for w in self.bus]
        return "".join([f", .{circuit_block.out.get_wire(self.bus.index(o)).prefix}({unique_out_wires.pop(unique_out_wires.index(o.prefix))})" if o.prefix in unique_out_wires else f", .{circuit_block.out.get_wire(self.bus.index(o)).prefix}()" for o in self.bus])

    def get_wire_declaration_v(self):
        """Declare the wire in Verilog code representation.

        Returns:
            str: Verilog code for declaration of individual bus wires.
        """
        return f"  wire [{self.N-1}:0] {self.prefix};\n"

    """ BLIF CODE GENERATION """
    def get_wire_declaration_blif(self):
        """Declare each wire from the bus independently in Blif code representation.

        Returns:
            str: Blif code for declaration of individual bus wires.
        """
        # Ensures correct binding between the bus wire index and the wire itself
        # It is used for the case when multiple of the same wire (e.g. `ContantWireValue0()`) are present in the bus (its id would otherwise be incorrect when using `self.bus.index(_)`)
        mapped_positions = [(w_id, self.bus[w_id]) for w_id in range(self.N)]
        array = True if self.N > 1 else False
        return "".join([f" {w[1].get_declaration_blif(prefix=self.prefix, offset=w[0], array=array)}" for w in mapped_positions])

    def get_wire_assign_blif(self, output: bool = False):
        """Assign all bits from the bus as each individual wires or assign wires into the corresponding output bus position in Blif code representation.

        Args:
            output (bool, optional): Specifies whether bus wires are used as outputs (True, assigned to) or as inputs (False, assigned from). Defaults to False.

        Returns:
            str: Blif code for bus wires assignments.
        """
        # Ensures correct binding between the bus wire index and the wire itself
        # It is used for the case when multiple of the same wire (e.g. `ContantWireValue0()`) are present in the bus (its id would otherwise be incorrect when using `self.bus.index(_)`)
        mapped_positions = [(w_id, self.bus[w_id]) for w_id in range(self.N)]
        if self.N > 1:
            return "".join([w[1].get_assign_blif(prefix=self.prefix+f"[{w[0]}]", output=output) for w in mapped_positions])
        else:
            return "".join([w[1].get_assign_blif(prefix=self.prefix, output=output) for w in mapped_positions])

    def get_unique_assign_out_wires_blif(self, function_block_out_bus: object):
        """Assigns unique output wires to their respective outputs of subcomponent's function block modul in hierarchical Blif subcomponent's invocation.

        Args:
            function_block_out_bus (object): Specifies output bus of corresponding function block's outputs for proper subcomponent modul invocation.

        Returns:
            str: Blif code for proper subcomponent's function block invocation with respective output wires assignment.
        """
        unique_out_wires = []
        [unique_out_wires.append(w.prefix) if w.prefix not in unique_out_wires else None for w in self.bus]
        return "".join([f" {function_block_out_bus.get_wire(self.bus.index(o)).name}={unique_out_wires.pop(unique_out_wires.index(o.prefix))}" if o.prefix in unique_out_wires else "" for o in self.bus])



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    PGSumLogic
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    OrGate,
    XorGate
)


class UnsignedPGRippleCarryAdder(GeneralCircuit):
    """Class representing unsigned ripple carry adder with propagate/generate logic.

    Unsigned ripple carry adder with PG logic represents slightly different rca implementation
    of N-bit unsigned adder which is composed of N one bit three input P/G/Sum logic function blocks.

    Each of them generates the P/G signals and contains a XOR gate responsible
    for the creation of corresponding sum bit. Besides this the Group PG logic is composed
    of AND and OR gates (basically a GreyCell) for computing the prefixes.

    You can find more info in the book: CMOS VLSI Design

    ```
      B3 A3         B2 A2       B1 A1       B0 A0
      │  │          │  │        │  │        │  │
    ┌─▼──▼─┐      ┌─▼──▼─┐    ┌─▼──▼─┐    ┌─▼──▼─┐
    │  PG  │  C3  │  PG  │ C2 │  PG  │ C1 │  PG  │
    │  SUM │◄────┐│  SUM │◄──┐│  SUM │◄──┐│  SUM │◄──0
    │      │     ││      │   ││      │   ││      │
    └─┬──┬┬┘     │└─┬┬┬──┘   │└─┬┬┬──┘   │└─┬┬┬──┘
      │  ││G3P3S3│  │││G2P2S2│  │││G1P1S1│  │││G0P0S0
      │ ┌▼▼──────┴──▼▼▼──────┴──▼▼▼──────┴──▼▼▼──┐
      │ │          Group PG logic                │
      │ │                                        │
      │ └─┬───────┬──────────┬──────────┬────────┘
      │   │S3     │S2        │S1        │S0
    ┌─▼───▼───────▼──────────▼──────────▼────────┐
    │                Sum + Cout                  │
    │                  logic                     │
    └┬────┬───────┬──────────┬──────────┬────────┘
     │    │       │          │          │
     ▼    ▼       ▼          ▼          ▼
    Cout  S3      S1         S0         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned P/G rca. Defaults to "".
        name (str, optional): Name of unsigned P/G rca. Defaults to "u_pg_rca".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_pg_rca", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # Gradual addition of 1-bit adder components
        for input_index in range(self.N):
            if input_index == 0:
                # First full adder with connected constant wire with value 0 as cin 0
                obj_pg_fa = PGSumLogic(self.a.get_wire(input_index), self.b.get_wire(input_index), ConstantWireValue0(), prefix=self.prefix+"_pg_sum"+str(input_index))
            else:
                obj_pg_fa = PGSumLogic(self.a.get_wire(input_index), self.b.get_wire(input_index), self.get_previous_component().out, prefix=self.prefix+"_pg_sum"+str(input_index))

            self.add_component(obj_pg_fa)
            self.out.connect(input_index, obj_pg_fa.get_sum_wire())

            obj_and = AndGate(self.get_previous_component().c, self.get_previous_component().get_propagate_wire(), prefix=self.prefix+"_and"+str(input_index), parent_component=self)
            obj_or = OrGate(obj_and.out, self.get_previous_component().get_generate_wire(), prefix=self.prefix+"_or"+str(input_index), parent_component=self)
            self.add_component(obj_and)
            self.add_component(obj_or)

            # Connecting last output bit to last cout
            if input_index == (self.N-1):
                self.out.connect(self.N, obj_or.out)


class SignedPGRippleCarryAdder(UnsignedPGRippleCarryAdder, GeneralCircuit):
    """Class representing signed ripple carry adder with propagate/generate logic.

    Signed ripple carry adder with PG logic represents slightly different rca implementation
    of N-bit signed adder which is composed of N one bit three input P/G/Sum logic function blocks.

    Each of them generates the P/G signals and contains a XOR gate responsible
    for the creation of corresponding sum bit. Besides this the Group PG logic is composed
    of AND and OR gates (basically a GreyCell) for computing the prefixes.

    At last XOR gates are used to ensure proper sign extension.

    You can find more info in the book: CMOS VLSI Design

    ```
      B3 A3         B2 A2       B1 A1       B0 A0
      │  │          │  │        │  │        │  │
    ┌─▼──▼─┐      ┌─▼──▼─┐    ┌─▼──▼─┐    ┌─▼──▼─┐
    │  PG  │  C3  │  PG  │ C2 │  PG  │ C1 │  PG  │
    │  SUM │◄────┐│  SUM │◄──┐│  SUM │◄──┐│  SUM │◄──0
    │      │     ││      │   ││      │   ││      │
    └─┬──┬┬┘     │└─┬┬┬──┘   │└─┬┬┬──┘   │└─┬┬┬──┘
      │  ││G3P3S3│  │││G2P2S2│  │││G1P1S1│  │││G0P0S0
      │ ┌▼▼──────┴──▼▼▼──────┴──▼▼▼──────┴──▼▼▼──┐
      │ │          Group PG logic                │
      │ │                                        │
      │ └─┬───────┬──────────┬──────────┬────────┘
      │   │S3     │S2        │S1        │S0
    ┌─▼───▼───────▼──────────▼──────────▼────────┐
    │            Sum + Cout logic                │
    │           with sign extension              │
    └┬────┬───────┬──────────┬──────────┬────────┘
     │    │       │          │          │
     ▼    ▼       ▼          ▼          ▼
    Cout  S3      S1         S0         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed P/G rca. Defaults to "".
        name (str, optional): Name of signed P/G rca. Defaults to "s_pg_rca".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_pg_rca", **kwargs):  
        # Signed bus length extension
        N = max(a.N, b.N)
        a.bus_extend(N=N, prefix=a.prefix, desired_extension_wire=a.get_wire(a.N-1))
        b.bus_extend(N=N, prefix=b.prefix, desired_extension_wire=b.get_wire(b.N-1))
    

        super().__init__(a=a, b=b, prefix=prefix, name=name, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        sign_xor_1 = XorGate(self.a.get_wire(self.N-1), self.b.get_wire(self.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(sign_xor_1)
        sign_xor_2 = XorGate(sign_xor_1.out, self.get_previous_component(2).out, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(sign_xor_2)
        self.out.connect(self.N, sign_xor_2.out)



"""
Testing the QuAdder
"""
import os
import sys
# Add the parent directory to the system path
DIR_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(DIR_PATH, '..'))
import numpy as np
import itertools
from ariths_gen.wire_components import Bus
from ariths_gen.multi_bit_circuits.approximate_adders import QuAdder


def test_quadder():
    c = QuAdder(Bus("a", 8), Bus("b", 8), R=[4, 2, 2], P=[0, 2, 2], prefix="quad")
    c.get_v_code_hier(file_object=sys.stdout)

    x = np.arange(0, 256).reshape(-1, 1)
    y = np.arange(0, 256).reshape(1, -1)

    r = c(x, y)
    r2 = x + y

    assert np.abs(r - r2).max() == 64
    np.testing.assert_equal(np.abs(r - r2).mean(), 7.5)


if __name__ == "__main__":
    test_quadder()
    print("Quadder Python tests were successful!")



from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    NandGate,
    OrGate,
    NorGate,
    XorGate,
    XnorGate,
    NotGate
)
import re


class UnsignedCGPCircuit(GeneralCircuit):
    """Unsigned circuit variant that loads CGP code and is able to export it to C/verilog/Blif/CGP."""

    def __init__(self, code: str = "", input_widths: list = None, inputs: list = None, prefix: str = "", name: str = "cgp", **kwargs):
        cgp_prefix, cgp_core, cgp_outputs = re.match(
            r"{(.*)}(.*)\(([^()]+)\)", code).groups()

        c_in, c_out, c_rows, c_cols, c_ni, c_no, c_lback = map(
            int, cgp_prefix.split(","))
        
        assert inputs is not None or input_widths is not None, "Either inputs or input_widths must be provided"

        if inputs:
            assert input_widths is None, "Only one of inputs or input_widths must be provided"

            input_widths =[i.N for i in inputs]
            assert sum(input_widths) == c_in, f"CGP input width {c_in} doesn't match inputs {inputs_widths}"



        else:

            assert sum(
                input_widths) == c_in, f"CGP input width {c_in} doesn't match input_widths {input_widths}"
            
            def get_name(i):
                if i < 26:
                    return chr(i + 0x61)
                assert i < 26 * 26
                return chr(i // 26 + 0x60) + chr(i % 26 + 0x61)

            inputs = [Bus(N=bw, prefix=f"input_{get_name(i)}")
                    for i, bw in enumerate(input_widths)]

            # Assign each Bus object in self.inputs to a named attribute of self
            for bus in inputs:
                # Here, bus.prefix is 'input_a', 'input_b', etc.
                # We strip 'input_' and use the remaining part (e.g., 'a', 'b') to create the attribute name
                attr_name = bus.prefix.replace('input_', '')
                setattr(self, attr_name, bus)

        # Adding values to the list
        self.vals = {}
        j = 2  # Start from two, 0=False, 1=True
        for iid, bw in enumerate(input_widths):
            for i in range(bw):
                assert j not in self.vals
                self.vals[j] = inputs[iid].get_wire(i)
                j += 1

        super().__init__(prefix=prefix, name=name, out_N=c_out, inputs=inputs, **kwargs)

        if not code:
            return # only for getting the name
        
        cgp_core = cgp_core.split(")(")

        i = 0
        for definition in cgp_core:
            i, in_a, in_b, fn = map(int, re.match(
                r"\(?\[(\d+)\](\d+),(\d+),(\d+)\)?", definition).groups())

            if in_a > i or in_b > i:
                raise ValueError(f"Backward connection in CGP gene \"{definition}\", maxid = {i}")

            if in_a == i or in_b == i:
                raise ValueError(f"Loop connection in CGP gene: \"{definition}\", maxid = {i}")

            comp_set = dict(prefix=f"{self.prefix}_core_{i:03d}", parent_component=self)

            a, b = self._get_wire(in_a), self._get_wire(in_b)
            if fn == 0:  # IDENTITY
                o = a
            elif fn == 1:  # NOT
                o = self.add_component(NotGate(a, **comp_set)).out
            elif fn == 2:  # AND
                o = self.add_component(AndGate(a, b, **comp_set)).out
            elif fn == 3:  # OR
                o = self.add_component(OrGate(a, b, **comp_set)).out
            elif fn == 4:  # XOR
                o = self.add_component(XorGate(a, b, **comp_set)).out
            elif fn == 5:  # NAND
                o = self.add_component(NandGate(a, b, **comp_set)).out
            elif fn == 6:  # NOR
                o = self.add_component(NorGate(a, b, **comp_set)).out
            elif fn == 7:  # XNOR
                o = self.add_component(XnorGate(a, b, **comp_set)).out
            elif fn == 8:  # TRUE
                o = ConstantWireValue1()
            elif fn == 9:  # FALSE
                o = ConstantWireValue0()

            assert i not in self.vals
            self.vals[i] = o

        # Output connection
        for i, o in enumerate(map(int, cgp_outputs.split(","))):
            if o >= c_in + c_rows * c_cols + 2:
                raise ValueError(
                    f"Output {i} is connected to wire {o} which is not in the range of CGP wires ({c_in + c_rows * c_cols + 2})")
            w = self._get_wire(o)
            self.out.connect(i, w)

    @staticmethod
    def get_inputs_outputs(code: str):
        cgp_prefix, cgp_core, cgp_outputs = re.match(
            r"{(.*)}(.*)\(([^()]+)\)", code).groups()

        c_in, c_out, c_rows, c_cols, c_ni, c_no, c_lback = map(
            int, cgp_prefix.split(","))

        return c_in, c_out

    def _get_wire(self, i):
        if i == 0:
            return ConstantWireValue0()
        if i == 1:
            return ConstantWireValue1()
        try:
            return self.vals[i]
        except KeyError:

            raise KeyError(f"Key {i} not found in " + ", ".join(
                [f"{i}: {v}" for i, v in self.vals.items()]
            ))


class SignedCGPCircuit(UnsignedCGPCircuit):
    """Signed circuit variant that loads CGP code and is able to export it to C/verilog/Blif/CGP."""
    def __init__(self, code: str, input_widths: list = None, inputs: list=None, prefix: str = "", name: str = "cgp", **kwargs):
        super().__init__(code=code, input_widths=input_widths, inputs=inputs, prefix=prefix, name=name, signed=True, **kwargs)
        self.c_data_type = "int64_t"



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    PGSumLogic,
    GreyCell,
    BlackCell
)
from ariths_gen.one_bit_circuits.logic_gates import (
    XorGate
)
import math


class UnsignedLadnerFischerAdder(GeneralCircuit):
    """Class representing unsigned Ladner-Fischer adder (using valency-2 logic gates).

    The Ladner-Fischer adder belongs to a type of tree (parallel-prefix) adders.
    Tree adder structure consists of three parts of logic: 1) PG logic generation, 2) Parallel PG logic computation, 3) Final sum and cout computation
    The main difference between each tree adder lies in the implementation of the part 2).

    Ladner-Fischer adders are a family of tree adders that represent a tradeoff between Brent-Kung and Sklansky implementations.
    Depending on the input bitwidth, there are many possible implementation configurations, precisely: [1, ⌈log2(N)⌉-2] number for N > 4 (otherwise just 1).
    The structures of the individual configurations shift from inclination more towards one or the other original implementation.

    Ladner-Fischer networks provide tradeoff between the number of stages in the PG logic and fanout load on wires.

    Main building components are GreyCells and BlackCells that appropriately encapsulate the essential logic used for PG computation.
    For further circuit characteristics see the book CMOS VLSI Design.

    The implementation performs the 1) and 3) (sum XORs) parts using one bit three input P/G/Sum logic function blocks.
    The 2) part is then composed according to the parallel-prefix adder characteristics.

    ```
      B3 A3         B2 A2       B1 A1       B0 A0
      │  │          │  │        │  │        │  │
    ┌─▼──▼─┐      ┌─▼──▼─┐    ┌─▼──▼─┐    ┌─▼──▼─┐
    │  PG  │  C3  │  PG  │ C2 │  PG  │ C1 │  PG  │
    │  SUM │◄────┐│  SUM │◄──┐│  SUM │◄──┐│  SUM │◄──0
    │      │     ││      │   ││      │   ││      │
    └─┬──┬┬┘     │└─┬┬┬──┘   │└─┬┬┬──┘   │└─┬┬┬──┘
      │  ││G3P3S3│  │││G2P2S2│  │││G1P1S1│  │││G0P0S0
      │ ┌▼▼──────┴──▼▼▼──────┴──▼▼▼──────┴──▼▼▼──┐
      │ │            Parallel-prefix             │
      │ │               PG logic                 │
      │ └─┬───────┬──────────┬──────────┬────────┘
      │   │S3     │S2        │S1        │S0
    ┌─▼───▼───────▼──────────▼──────────▼────────┐
    │                Sum + Cout                  │
    │                   logic                    │
    └┬────┬───────┬──────────┬──────────┬────────┘
     │    │       │          │          │
     ▼    ▼       ▼          ▼          ▼
    Cout  S3      S1         S0         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned lfa. Defaults to "".
        name (str, optional): Name of unsigned lfa. Defaults to "u_lfa".
        config_choice (int, optional): Tradeoff implementation choice concerning the number of stages in the PG logic and fanout load on wires. The number of choices goes from 1 up to ⌈log2(N)⌉-2 for N > 4, otherwise the choice is 1. Defaults to 1.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_lfa", config_choice: int = 1, **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)
        cin = ConstantWireValue0()

        # Configuration setting
        self.config_choice = config_choice
        if self.N > 4:
            assert self.config_choice > 0 and self.config_choice <= math.ceil(math.log(self.N, 2))-2, "The configuration choice must fall in a range [1, ⌈log2(N)⌉-2] for N > 4."
        else:
            assert self.config_choice == 1, "The configuration choice for N <= 4 is only 1."

        # Lists of list containing all propagate/generate wires
        self.propagate_sig = []
        self.generate_sig = []
        # Cin0 used as a first generate wire for obtaining next carry bits
        self.generate_sig.append([cin])
        # The configuration choice offset used to appropriately interconnect PG logic cells in each stage and to determine the number of stages for a given bit index
        config_offset = 2**self.config_choice

        # For each bit pair proceed with three stages of PPAs (Generate PG signals, Prefix computation (PG logic), Computation and connection of outputs)
        # NOTE In the implementation below, both the first and third stages are handled by the PG FAs
        for i_wire in range(self.N):
            # 1st + 3rd stage: Generate PG signals + Computation and connection of outputs
            self.add_component(PGSumLogic(self.a.get_wire(i_wire), self.b.get_wire(i_wire), self.generate_sig[i_wire][-1], prefix=self.prefix+"_pg_sum"+str(self.get_instance_num(cls=PGSumLogic))))
            self.generate_sig.append([self.get_previous_component().get_generate_wire()])
            self.propagate_sig.append([self.get_previous_component().get_propagate_wire()])

            self.out.connect(i_wire, self.get_previous_component().get_sum_wire())
            if i_wire == self.N-1:
                self.add_component(GreyCell(self.generate_sig[i_wire+1][-1], self.propagate_sig[i_wire][0], self.generate_sig[i_wire][-1], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                self.out.connect(self.N, self.get_previous_component().get_generate_wire())
            # 2nd stage: Prefix Computation (PG logic)
            # For all bit indexes expect for the last one, proceed with the parralel prefix PG logic
            else:
                if (i_wire+2) % config_offset == 0 or (i_wire < 4 and i_wire % 2 == 0):  # These input bits form a Sklansky adder structure
                    binary_form = bin(i_wire+1)[2:][::-1]  # +1 to obtain proper carry bit index (index 0 corresponds to Cin)
                    index_stages = len(binary_form)
                    prev_stage_int_value = 0
                    for stage, value in enumerate(binary_form):
                        if value == '1':
                            # Grey cell
                            if stage == index_stages-1:
                                if index_stages == 1:  # Bit index with only one stage
                                    self.add_component(GreyCell(self.generate_sig[i_wire+1][0], self.propagate_sig[i_wire][0], self.generate_sig[i_wire][-1], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                                else:  # Bit index contains multiple stages, GC is its last stage
                                    self.add_component(GreyCell(self.get_previous_component().get_generate_wire(), self.get_previous_component().get_propagate_wire(), self.generate_sig[i_wire-prev_stage_int_value][-1], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                            # Black cell
                            else:  # If bit index contains more than one stage, every stage except for the last one contains a Black cell
                                self.add_component(BlackCell(self.generate_sig[i_wire+1][-1], self.propagate_sig[i_wire][-1], self.generate_sig[i_wire-prev_stage_int_value][stage], self.propagate_sig[i_wire-1-prev_stage_int_value][stage], prefix=self.prefix+"_bc"+str(self.get_instance_num(cls=BlackCell))))
                                self.propagate_sig[i_wire].append(self.get_previous_component().get_propagate_wire())
                            self.generate_sig[i_wire+1].append(self.get_previous_component().get_generate_wire())
                            prev_stage_int_value += 2**stage
                else:  # Other input bits are implemented in the same fasion as in the Brent-Kung adder structure
                    index_stages = math.log(i_wire+2, 2)  # +1 because indexes start from 0 and additional +1 because the first generated carry is actually the second carry after cin (the previous cin has 0 stages)
                    if int(index_stages) == index_stages:  # Bit indexes that are powers of 2
                        for stage in range(int(index_stages)):
                            # Grey cell
                            if stage == index_stages-1:
                                if stage == 0:  # Bit index with only one stage
                                    self.add_component(GreyCell(self.generate_sig[i_wire+1][0], self.propagate_sig[i_wire][0], self.generate_sig[i_wire][0], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                                else:  # Bit index contains multiple stages, GC is its last stage
                                    self.add_component(GreyCell(self.get_previous_component().get_generate_wire(), self.get_previous_component().get_propagate_wire(), self.generate_sig[(i_wire+1)-(2**stage)][-1], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                            # Black cell
                            else:  # If bit index contains more than one stage, every stage except for the last one contains a Black cell
                                self.add_component(BlackCell(self.generate_sig[i_wire+1][-1], self.propagate_sig[i_wire][-1], self.generate_sig[(i_wire+1)-(2**stage)][stage], self.propagate_sig[i_wire-(2**stage)][stage], prefix=self.prefix+"_bc"+str(self.get_instance_num(cls=BlackCell))))
                                self.propagate_sig[i_wire].append(self.get_previous_component().get_propagate_wire())
                            self.generate_sig[i_wire+1].append(self.get_previous_component().get_generate_wire())
                    elif i_wire % 2 == 0:  # Even-numbered input indexes
                        # Different handling based on the chosen configuration and the corresponding PG logic structure
                        if i_wire+2 <= config_offset:  # Same behaviour as in the case of a regular Brent-Kung
                            lower_closest_power_of_two = 2**math.floor(math.log(i_wire+2, 2))  # i.e. 2^3 = 8
                            higher_closest_power_of_two = 2**math.ceil(math.log(i_wire+2, 2))  # i.e. 2^4 = 16
                            down_diff = (i_wire+2) - lower_closest_power_of_two
                            up_diff = higher_closest_power_of_two - (i_wire+2)
                        else:  # The number of stages for these indexes is different as opposed to a regular Brent-Kung implementation
                            # Determine the closest lower and higher multiples of the input config offset
                            lower_closest_multiple = ((i_wire+2) // config_offset) * config_offset
                            higher_closest_multiple = lower_closest_multiple + config_offset
                            down_diff = abs((i_wire+2) - lower_closest_multiple)
                            up_diff = abs(higher_closest_multiple - (i_wire+2))

                        diff_power = up_diff if up_diff < down_diff else down_diff
                        index_stages = math.floor(math.log(diff_power, 2))+1
                        for stage in range(index_stages):
                            # Black cell
                            if stage != index_stages-1:  # Bit index contains more than one stage, every stage except for the last one contains a Black cell
                                self.add_component(BlackCell(self.generate_sig[i_wire+1][-1], self.propagate_sig[i_wire][-1], self.generate_sig[(i_wire+1)-(2**stage)][stage], self.propagate_sig[i_wire-(2**stage)][stage], prefix=self.prefix+"_bc"+str(self.get_instance_num(cls=BlackCell))))
                                self.propagate_sig[i_wire].append(self.get_previous_component().get_propagate_wire())
                            # Grey cell
                            else:  # Bit index contains multiple stages, GC is its last stage
                                self.add_component(GreyCell(self.get_previous_component().get_generate_wire(), self.get_previous_component().get_propagate_wire(), self.generate_sig[(i_wire+1)-(2**stage)][-1], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                            self.generate_sig[i_wire+1].append(self.get_previous_component().get_generate_wire())
                    else:  # Odd-numbered input indexes
                        self.add_component(GreyCell(self.generate_sig[i_wire+1][-1], self.propagate_sig[i_wire][0], self.generate_sig[i_wire][-1], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                        self.generate_sig[i_wire+1].append(self.get_previous_component().get_generate_wire())


class SignedLadnerFischerAdder(UnsignedLadnerFischerAdder, GeneralCircuit):
    """Class representing signed Ladner-Fischer adder (using valency-2 logic gates).

    The Ladner-Fischer adder belongs to a type of tree (parallel-prefix) adders.
    Tree adder structure consists of three parts of logic: 1) PG logic generation, 2) Parallel PG logic computation, 3) Final sum and cout computation
    The main difference between each tree adder lies in the implementation of the part 2).

    Ladner-Fischer adders are a family of tree adders that represent a tradeoff between Brent-Kung and Sklansky implementations.
    Depending on the input bitwidth, there are many possible implementation configurations, precisely: [1, ⌈log2(N)⌉-2] number for N > 4 (otherwise just 1).
    The structures of the individual configurations shift from inclination more towards one or the other original implementation.

    Ladner-Fischer networks provide tradeoff between the number of stages in the PG logic and fanout load on wires.

    Main building components are GreyCells and BlackCells that appropriately encapsulate the essential logic used for PG computation.
    For further circuit characteristics see the book CMOS VLSI Design.

    The implementation performs the 1) and 3) (sum XORs) parts using one bit three input P/G/Sum logic function blocks.
    The 2) part is then composed according to the parallel-prefix adder characteristics.
    At last XOR gates are used to ensure proper sign extension.

    ```
      B3 A3         B2 A2       B1 A1       B0 A0
      │  │          │  │        │  │        │  │
    ┌─▼──▼─┐      ┌─▼──▼─┐    ┌─▼──▼─┐    ┌─▼──▼─┐
    │  PG  │  C3  │  PG  │ C2 │  PG  │ C1 │  PG  │
    │  SUM │◄────┐│  SUM │◄──┐│  SUM │◄──┐│  SUM │◄──0
    │      │     ││      │   ││      │   ││      │
    └─┬──┬┬┘     │└─┬┬┬──┘   │└─┬┬┬──┘   │└─┬┬┬──┘
      │  ││G3P3S3│  │││G2P2S2│  │││G1P1S1│  │││G0P0S0
      │ ┌▼▼──────┴──▼▼▼──────┴──▼▼▼──────┴──▼▼▼──┐
      │ │            Parallel-prefix             │
      │ │               PG logic                 │
      │ └─┬───────┬──────────┬──────────┬────────┘
      │   │S3     │S2        │S1        │S0
    ┌─▼───▼───────▼──────────▼──────────▼────────┐
    │                Sum + Cout                  │
    │           with sign extension              │
    └┬────┬───────┬──────────┬──────────┬────────┘
     │    │       │          │          │
     ▼    ▼       ▼          ▼          ▼
    Cout  S3      S1         S0         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed lfa. Defaults to "".
        name (str, optional): Name of signed lfa. Defaults to "s_lfa".
        config_choice (int, optional): Tradeoff implementation choice concerning the number of stages in the PG logic and fanout load on wires. The number of choices goes from 1 up to ⌈log2(N)⌉-2 for N > 4, otherwise the choice is 1. Defaults to 1.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_lfa", config_choice: int = 1, **kwargs):
        # Signed bus length extension
        N = max(a.N, b.N)
        a.bus_extend(N=N, prefix=a.prefix, desired_extension_wire=a.get_wire(a.N-1))
        b.bus_extend(N=N, prefix=b.prefix, desired_extension_wire=b.get_wire(b.N-1))
        
        super().__init__(a=a, b=b, prefix=prefix, name=name, config_choice=config_choice, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        self.add_component(XorGate(self.a.get_wire(self.N-1), self.b.get_wire(self.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self))
        self.add_component(XorGate(self.get_previous_component().out, self.get_previous_component(2).get_generate_wire(), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self))
        self.out.connect(self.N, self.get_previous_component().out)



import re

class Wire():
    """Class representing basic wire used to interconnect components.

    Description of the __init__ method.

    Args:
        name (str): Name of the wire.
        prefix (str, optional): Prefix of the wire. Defaultly the same as its name. Defaults to "".
        value (int, optional): Value the wire carries (0,1). Defaults to 0.
        index (int, optional): Index position of wire (mainly used for indexing within a bus). Defaults to 0.
        parent_bus (object, optional): Bus object of which a Wire is a part of (used mainly for proper generation of wire names). Defaults to None.
    """
    def __init__(self, name: str, prefix: str = "", value: int = 0, index: int = 0, parent_bus: object = None):
        self.name = name
        self.value = value
        self.index = index
        self.prefix = name if prefix == "" else prefix
        self.parent_bus = parent_bus

    def __str__(self):
        return f"wire{self.name}{self.value}{self.index}"

    @staticmethod
    def is_const():
        """Information whether wire carries constant value.

        Returns:
            bool: False, because basic wire doesn't represent a wire with constant value.
        """
        return False

    def is_buswire(self):
        return self.name.endswith("["+str(self.index)+"]") and self.parent_bus is not None

    """ PYTHON CODE GENERATION """
    def return_wire_value_python_flat(self, offset: int = 0):
        """Retrieves desired bit value from wire represented in Python code variable (object) and bitwise shifts it to desired position for storing it within a bus for flat generation.

        Args:
            offset (int, optional): Used to shift wire value in order to be stored in proper location inside a bus. Defaults to 0.

        Returns:
            str: Python code bitwise shift for storing (constant/variable) wire value at desired offset position.
        """
        if self.is_const():
            return f"(({self.c_const}) << {offset})\n"
        # If wire is part of an input bus (where wire names are concatenated from bus prefix and their index position inside the bus in square brackets)
        # then the wire value is obtained from bitwise shifting the required wire from the parent bus ('parent_bus.prefix' is the same value as 'self.prefix')
        elif self.is_buswire():
            return f"((({self.prefix} >> {self.index}) & 0x01) << {offset})\n"

        else:
            return f"((({self.name} >> 0) & 0x01) << {offset})\n"

    """ C CODE GENERATION """
    def get_declaration_c(self):
        """Wire declaration in C code.

        Returns:
            str: Empty string if C code wire is carrying constant value (constant value is used in situ) or returns C code for declaration and initialization of wire's name.
        """
        if self.is_const():
            return ""
        else:
            return f"  uint8_t {self.name} = {self.value};\n"

    def get_wire_value_c_flat(self):
        """Accesses desired bit value from wire represented in C code variable used for flat generation.

        Returns:
            str: C code bitwise shift to get desired bit value from this wire or wire variable's constant bit value 0/1.
        """
        if self.is_const():
            return f"({self.c_const})"
        # If wire is part of an input bus (where wire names are concatenated from bus prefix and their index position inside the bus in square brackets)
        # then the wire value is obtained from bitwise shifting the required wire from the parent bus ('parent_bus.prefix' is the same value as 'self.prefix')
        elif self.is_buswire() and self.name == f"{self.prefix}[{self.index}]":
            return f"(({self.prefix} >> {self.index}) & 0x01)"
        elif self.is_buswire():
            g = re.match(r"(.*)\[(\d+)\]", self.name)
            if g:
                return f"(({g.group(1)} >> {g.group(2)}) & 0x01)"
            else:
                return f"(({self.name} >> 0) & 0x01)"
        else:
            return f"(({self.name} >> 0) & 0x01)"

    def get_wire_value_c_hier(self):
        """Accesses desired bit value from wire represented in C code variable used for hierarchical generation.

        Returns:
            str: C code bitwise shift to get desired bit value position from this wire or wire variable's constant bit value 0/1.
        """
        if self.is_const():
            return f"({self.c_const})"
        else:
            return f"(({self.prefix} >> {self.index}) & 0x01)"

    def return_wire_value_c_flat(self, offset: int = 0):
        """Retrieves desired bit value from wire represented in C code variable and bitwise shifts it to desired position for storing it within a bus for flat generation.

        Args:
            offset (int, optional): Used to shift wire value in order to be stored in proper location inside a bus. Defaults to 0.

        Returns:
            str: C code bitwise shift for storing (constant/variable) wire value at desired offset position.
        """
        if self.is_const():
            return f"({self.c_const}) << {offset};\n"
        else:
            return f"(({self.name} >> 0) & 0x01ull) << {offset};\n"

    def return_wire_value_c_hier(self, offset: int = 0):
        """Retrieves desired bit value from wire represented in C code variable and bitwise shifts it to desired position for storing it within a bus for hierarchical generation.

        Args:
            offset (int, optional): Used to shift wire value in order to be stored in proper location inside a bus. Defaults to 0.

        Returns:
            str: C code bitwise shift for storing (constant/variable) wire value at desired offset position.
        """
        # If wire is part of an input bus (where wire names are concatenated from bus prefix and their index position inside the bus in square brackets)
        # then the wire value is obtained from bitwise shifting the required wire from the parent bus ('parent_bus.prefix' is the same value as 'self.prefix')
        if self.is_const():
            return f"({self.c_const}) << {offset};\n"
        else:
            return f"(({self.prefix} >> {self.index}) & 0x01ull) << {offset};\n"

    """ VERILOG CODE GENERATION """
    def get_declaration_v_flat(self):
        """Wire declaration for flat Verilog code.

        Returns:
            str: Empty string if Verilog code wire is carrying constant value (constant value is used in situ) or returns Verilog code for declaration and initialization of wire's name.
        """
        if self.is_const():
            return ""
        else:
            return f"  wire {self.name};\n"

    def get_declaration_v_hier(self):
        """Wire declaration for hierarchical Verilog code.

        Returns:
            str: Empty string if Verilog code wire is carrying constant value (constant value is used in situ) or returns Verilog code for declaration and initialization of wire's name.
        """
        if self.is_const():
            return ""
        else:
            return f"  wire [0:0] {self.name};\n"

    def get_wire_value_v_flat(self):
        """Accesses bit value from wire represented in Verilog code variable used for flat generation.

        Returns:
            str: Verilog code to get bit value from this wire or to get constant wire's bit value 0/1.
        """
        if self.is_const():
            return self.v_const
        else:
            return self.name

    def get_wire_value_v_hier(self):
        """Accesses bit value from wire represented in Verilog code variable used for hierarchical generation.

        Returns:
            str: Verilog code to get bit value from this wire or to get constant wire's bit value 0/1.
        """
        if self.is_const():
            return self.v_const
        else:
            return f"{self.prefix}[{self.index}]"

    def return_wire_value_v_flat(self):
        """Retrieves bit value from wire represented in Verilog code variable for storing it within a bus for flat generation.

        Returns:
            str: Verilog code for retrieving (constant/variable) wire value (and assign it at desired bus offset position).
        """
        if self.is_const():
            return f"{self.v_const};\n"
        else:
            return f"{self.name};\n"

    def return_wire_value_v_hier(self):
        """Retrieves bit value from bus's wires and stores them in bus's corresponding Verilog variable at proper offset bit position in the bus for hierarchical generation.

        Returns:
            str: Verilog code for retrieving (constant/variable) wire value used for assigning it into bus represented in Verilog code variable.
        """
        if self.is_const():
            return f"{self.v_const};\n"
        else:
            return f"{self.prefix}[{self.index}];\n"

    """ BLIF CODE GENERATION """
    def get_declaration_blif(self, prefix: str = "", offset: int = 0, array: bool = False):
        """Declaration of wire which is part of a bus in Blif code.

        Declares basic wire name if wire is not part of a bus
        or declares wire by an offset of its position within the bus.

        Args:
            prefix (str, optional): Bus prefix of which this wire is a part off. Defaults to "".
            offset (int, optional): Offset wire location within a bus. Defaults to 0.
            array (bool, optional): Tells whether a basic wire or a wire from within a bus is to be declared. Defaults to False.

        Returns:
            str: Blif code for declaration of a wire.
        """
        if array is True:
            return f"{prefix}[{offset}]"
        else:
            return f"{prefix}"

    def get_wire_declaration_blif(self):
        """Declaration of a single wire in Blif code.
        
        Used for declaration of modul inputs.
        
        Returns:
            str: Blif code for declaration of a wire.
        """
        return f"{self.prefix} "
        
    def get_assign_blif(self, prefix: str, output: bool = False):
        """Assignment of wire value to another desired wire in Blif code.

        This wire's value is either assigned to desired output bus wire (represented by `prefix` name) when `output`=True.
        Otherwise the wire value at desired bus position (represented by `prefix` name) is assigned to this wire when `output`=False.

        Args:
            prefix (str): Name of the source/destination bus wire to be assigned from/to.
            output (bool, optional): Whether `prefix` represents the destination or the source wire in the assignment. Defaultly it symbolizes the source. Defaults to False.

        Returns:
            str: Blif code for assignment of one wire to another.
        """
        if output is True:
            if self.is_const():
                return f".names {self.blif_const} {prefix}\n" + \
                       f"1 1\n"
            else:
                return f".names {self.name} {prefix}\n" + \
                       f"1 1\n"
        else:
            if self.is_const():
                return "\n"
            else:
                return f".names {prefix} {self.name}\n" + \
                       f"1 1\n"

    def get_wire_value_blif(self):
        """Accesses bit value from wire represented in Blif code.

        Used for assignment of specific one bit circuit/gate values to their respective parameters
        in hierarchical Blif subcomponents generation.

        Returns:
            str: Blif code to get bit value from this wire or to get constant wire's bit value 0/1.
        """
        if self.is_const():
            return self.blif_const
        else:
            return self.name

    def __str__(self):
        if self.is_const():
            return f"<w={self.c_const}>"
        elif self.is_buswire():
            return f"<w={self.prefix}[{self.index}]>"
        else:
            return f"<w={self.name}>"

    """ define read-only parameter N"""
    @property
    def N(self):
        return 1


# Wires with constant values #
class ConstantWireValue0(Wire):
    """Class representing wire carrying constant value 0 used to interconnect components.

    Description of the __init__ method.

    Method fills in desired constant wire's attributes regarding its values for individual representations.

    Args:
        name (str, optional): Custom constant wire name and prefix (used for generation of circuits that use constants as inputs). Defaults to "".
    """
    def __init__(self, name: str = ""):
        self.name = "constant_value_0" if name == "" else name
        self.prefix = "constant_value_0" if name == "" else name
        self.index = 0
        self.value = 0
        self.parent_bus = None

        self.c_const = "0x00"
        self.v_const = "1'b0"
        self.blif_const = "gnd"
        # Constant wire id for CGP generation
        self.cgp_const = 0

    @staticmethod
    def is_const():
        """Information whether wire carries constant value.

        Returns:
            bool: True, because constant wire carries a constant value 0.
        """
        return True


class ConstantWireValue1(Wire):
    """Class representing wire carrying constant value 1 used to interconnect components.

    Description of the __init__ method.

    Method fills in desired constant wire's attributes regarding its values for individual representations.

    Args:
        name (str, optional): Custom constant wire name and prefix (used for generation of circuits that use constants as inputs). Defaults to "".
    """
    def __init__(self, name: str = ""):
        self.name = "constant_value_1" if name == "" else name
        self.prefix = "constant_value_1" if name == "" else name
        self.index = 0
        self.value = 1
        self.parent_bus = None

        self.c_const = "0x01"
        self.v_const = "1'b1"
        self.blif_const = "vdd"
        # Constant wire id for CGP generation
        self.cgp_const = 1

    @staticmethod
    def is_const():
        """Information whether wire carries constant value.

        Returns:
            bool: True, because constant wire carries a constant value 1.
        """
        return True



import os
import sys
# Add the parent directory to the system path
DIR_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(DIR_PATH, '..'))
import numpy as np
import math

from ariths_gen.wire_components import (
    Wire,
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)

from ariths_gen.core.arithmetic_circuits import GeneralCircuit

from ariths_gen.multi_bit_circuits.adders import (
    UnsignedCarryLookaheadAdder,
    UnsignedPGRippleCarryAdder,
    UnsignedRippleCarryAdder,
    SignedCarryLookaheadAdder,
    SignedPGRippleCarryAdder,
    SignedRippleCarryAdder,
    UnsignedCarrySkipAdder,
    SignedCarrySkipAdder,
    UnsignedKoggeStoneAdder,
    SignedKoggeStoneAdder,
    UnsignedBrentKungAdder,
    SignedBrentKungAdder,
    UnsignedSklanskyAdder,
    SignedSklanskyAdder,
    UnsignedHanCarlsonAdder,
    SignedHanCarlsonAdder,
    UnsignedLadnerFischerAdder,
    SignedLadnerFischerAdder,
    UnsignedKnowlesAdder,
    SignedKnowlesAdder,
    UnsignedCarrySelectAdder,
    SignedCarrySelectAdder,
    UnsignedConditionalSumAdder,
    SignedConditionalSumAdder,
    UnsignedCarryIncrementAdder,
    SignedCarryIncrementAdder
)

from ariths_gen.multi_bit_circuits.subtractors import (
    UnsignedRippleCarrySubtractor, SignedRippleCarrySubtractor,
    UnsignedRippleBorrowSubtractor, SignedRippleBorrowSubtractor
)

from ariths_gen.multi_bit_circuits.multipliers import (
    UnsignedDaddaMultiplier,
    UnsignedArrayMultiplier,
    UnsignedWallaceMultiplier,
    UnsignedCarrySaveMultiplier,
    SignedArrayMultiplier,
    SignedDaddaMultiplier,
    SignedWallaceMultiplier,
    SignedCarrySaveMultiplier
)

from ariths_gen.multi_bit_circuits.approximate_multipliers import (
    UnsignedTruncatedArrayMultiplier,
    UnsignedTruncatedCarrySaveMultiplier,
    UnsignedBrokenArrayMultiplier,
    UnsignedBrokenCarrySaveMultiplier,
    UnsignedRecursiveMultiplier,
    UnsignedAccurateTwoBitMultiplier
)

from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    NandGate,
    OrGate,
    NorGate,
    XorGate,
    XnorGate,
    NotGate
)


def test_unsigned_approxmul(values=False):
    """ Test unsigned approximate multipliers """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N)
    bv = av.reshape(-1, 1)
    expected = av * bv

    for c in [UnsignedBrokenArrayMultiplier, UnsignedBrokenCarrySaveMultiplier, UnsignedTruncatedArrayMultiplier, UnsignedTruncatedCarrySaveMultiplier]:
        if c == UnsignedTruncatedArrayMultiplier or c == UnsignedTruncatedCarrySaveMultiplier:
            mul = c(a=a, b=b, truncation_cut=2)
        elif c == UnsignedBrokenArrayMultiplier or c == UnsignedBrokenCarrySaveMultiplier:
            mul = c(a=a, b=b, horizontal_cut=1, vertical_cut=2)
    r = mul(av, bv)

    # WCE – worst case error; used for approximate multiplier error measurement
    WCE = np.amax(abs(np.subtract(r, expected)))

    # WCRE – worst case relative error; used for approximate multiplier error measurement
    np.seterr(divide='ignore', invalid='ignore')
    WCRE = np.max(np.nan_to_num(abs(np.subtract(r, expected)) / expected))

    if isinstance(mul, UnsignedTruncatedArrayMultiplier) or isinstance(mul, UnsignedTruncatedCarrySaveMultiplier):
        # WCE_TM(n,k) = (2^k - 1) * (2^(n+1) - 2^k - 1)
        expected_WCE = (2 ** mul.truncation_cut - 1) * (2 ** (mul.a.N+1) - 2 ** mul.truncation_cut - 1)
    elif isinstance(mul, UnsignedBrokenArrayMultiplier) or isinstance(mul, UnsignedBrokenCarrySaveMultiplier):
        # WCE_BAM(n,h,v) = (2^n - 1) * {SUM_i0_to_h-1}(2^i) + 2^h * {SUM_i0_to_v-h-1}(2^(v-h) - 2^i)
        sum_1 = sum([2**i for i in range(0, mul.horizontal_cut)])
        sum_2 = sum([2**(mul.vertical_cut-mul.horizontal_cut) - 2**i for i in range(0, mul.vertical_cut-mul.horizontal_cut)])
        expected_WCE = (2 ** mul.N - 1) * sum_1 + 2 ** mul.horizontal_cut * sum_2

    # Test expected result
    assert expected_WCE == WCE
    if values is True:
        np.testing.assert_array_equal(expected, r)


def test_unsigned_mul():
    """ Test unsigned multipliers """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N)
    bv = av.reshape(-1, 1)
    expected = av * bv

    # No configurability
    for c in [UnsignedArrayMultiplier]:
        mul = c(a, b)
        assert mul(0, 0) == 0
        r = mul(av, bv)
        np.testing.assert_array_equal(expected, r)
    
    # Accurate variant of recursive multiplier
    for c in [UnsignedRecursiveMultiplier]:
        N_rec = 8
        a_rec = Bus(N=N_rec, prefix="a")
        b_rec = Bus(N=N_rec, prefix="b")
        av_rec = np.arange(2**N_rec)
        bv_rec = av_rec.reshape(-1, 1)
        expected_rec = av_rec * bv_rec
        mul = c(a_rec, b_rec, submultipliers=[UnsignedAccurateTwoBitMultiplier for _ in range((N_rec//2)**2)])
        assert mul(0, 0) == 0
        r = mul(av_rec, bv_rec)
        np.testing.assert_array_equal(expected_rec, r)

    # Configurable PPA
    for c in [UnsignedDaddaMultiplier, UnsignedCarrySaveMultiplier, UnsignedWallaceMultiplier]:
        # Non configurable multi-bit adders
        for ppa in [UnsignedPGRippleCarryAdder, UnsignedRippleCarryAdder, UnsignedConditionalSumAdder, UnsignedKoggeStoneAdder, UnsignedBrentKungAdder, UnsignedSklanskyAdder]:
            # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
            if c == UnsignedWallaceMultiplier:
                mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False)
                assert mul(0, 0) == 0
                r = mul(av, bv)
                np.testing.assert_array_equal(expected, r)

            mul = c(a, b, unsigned_adder_class_name=ppa)
            assert mul(0, 0) == 0
            r = mul(av, bv)
            np.testing.assert_array_equal(expected, r)

        # Multi-bit adders with configurable (uniform) logic blocks for parallel prefix computation (the ppa will use the block_size argument it recognizes, others are ignored)
        for ppa in [UnsignedCarryLookaheadAdder, UnsignedCarrySkipAdder, UnsignedCarrySelectAdder, UnsignedCarryIncrementAdder]:
            for i in range(1, N+1):
                # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
                if c == UnsignedWallaceMultiplier:
                    mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False, cla_block_size=i, bypass_block_size=i, select_block_size=i, increment_block_size=i)
                    assert mul(0, 0) == 0
                    r = mul(av, bv)
                    np.testing.assert_array_equal(expected, r)

                mul = c(a, b, unsigned_adder_class_name=ppa, cla_block_size=i, bypass_block_size=i, select_block_size=i, increment_block_size=i)
                assert mul(0, 0) == 0
                r = mul(av, bv)
                np.testing.assert_array_equal(expected, r)

        # Multi-bit tree adders with configurable structure based on input bit width (NOTE for showcase here, the second config would be applicable from bit width 9 onward; not tested here for the sake of saving deployment testing time)
        for adder in [UnsignedHanCarlsonAdder, UnsignedKnowlesAdder, UnsignedLadnerFischerAdder]:
            for i in range(1, N+1):
                # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
                if c == UnsignedWallaceMultiplier:
                    mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False, config_choice=1)
                    assert mul(0, 0) == 0
                    r = mul(av, bv)
                    np.testing.assert_array_equal(expected, r)

                mul = c(a, b, unsigned_adder_class_name=ppa, config_choice=1)
                assert mul(0, 0) == 0
                r = mul(av, bv)
                np.testing.assert_array_equal(expected, r)


def test_signed_mul():
    """ Test signed multipliers """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(-(2**(N-1)), 2**(N-1))
    bv = av.reshape(-1, 1)
    expected = av * bv

    # No configurability
    for c in [SignedArrayMultiplier]:
        mul = c(a, b)
        assert mul(0, 0) == 0
        r = mul(av, bv)
        np.testing.assert_array_equal(expected, r)

    # Configurable PPA
    for c in [SignedDaddaMultiplier, SignedCarrySaveMultiplier, SignedWallaceMultiplier]:
        # Non configurable multi-bit adders
        for ppa in [UnsignedPGRippleCarryAdder, UnsignedRippleCarryAdder, UnsignedConditionalSumAdder, UnsignedKoggeStoneAdder, UnsignedBrentKungAdder, UnsignedSklanskyAdder]:
            # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
            if c == UnsignedWallaceMultiplier:
                mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False)
                assert mul(0, 0) == 0
                r = mul(av, bv)
                np.testing.assert_array_equal(expected, r)

            mul = c(a, b, unsigned_adder_class_name=ppa)
            assert mul(0, 0) == 0
            r = mul(av, bv)
            np.testing.assert_array_equal(expected, r)

        # Multi-bit adders with configurable (uniform) logic blocks for parallel prefix computation (the ppa will use the block_size argument it recognizes, others are ignored)
        for ppa in [UnsignedCarryLookaheadAdder, UnsignedCarrySkipAdder, UnsignedCarrySelectAdder, UnsignedCarryIncrementAdder]:
            for bs in range(1, N+1):
                # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
                if c == UnsignedWallaceMultiplier:
                    mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False, cla_block_size=bs, bypass_block_size=bs, select_block_size=bs, increment_block_size=bs)
                    assert mul(0, 0) == 0
                    r = mul(av, bv)
                    np.testing.assert_array_equal(expected, r)

                mul = c(a, b, unsigned_adder_class_name=ppa, cla_block_size=bs, bypass_block_size=bs, select_block_size=bs, increment_block_size=bs)
                assert mul(0, 0) == 0
                r = mul(av, bv)
                np.testing.assert_array_equal(expected, r)

        # Multi-bit tree adders with configurable structure based on input bit width (NOTE for showcase here, the second config would be applicable from bit width 9 onward; not tested here for the sake of saving deployment testing time)
        for adder in [UnsignedHanCarlsonAdder, UnsignedKnowlesAdder, UnsignedLadnerFischerAdder]:
            for i in range(1, N+1):
                # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
                if c == UnsignedWallaceMultiplier:
                    mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False, config_choice=1)
                    assert mul(0, 0) == 0
                    r = mul(av, bv)
                    np.testing.assert_array_equal(expected, r)

                mul = c(a, b, unsigned_adder_class_name=ppa, config_choice=1)
                assert mul(0, 0) == 0
                r = mul(av, bv)
                np.testing.assert_array_equal(expected, r)


def test_unsigned_add():
    """ Test unsigned adders """
    N = 9
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N)
    bv = av.reshape(-1, 1)
    expected = av + bv

    # Non configurable multi-bit adders
    for c in [UnsignedPGRippleCarryAdder, UnsignedRippleCarryAdder, UnsignedConditionalSumAdder, UnsignedKoggeStoneAdder, UnsignedBrentKungAdder, UnsignedSklanskyAdder]:
        add = c(a, b)
        r = add(av, bv)
        np.testing.assert_array_equal(expected, r)

    # Multi-bit adders with configurable (uniform) logic blocks for parallel prefix computation (the adder will use the block_size argument it recognizes, others are ignored)
    for c in [UnsignedCarryLookaheadAdder, UnsignedCarrySkipAdder, UnsignedCarrySelectAdder, UnsignedCarryIncrementAdder]:
        for bs in range(1, N+1):
            add = c(a, b, cla_block_size=bs, bypass_block_size=bs, select_block_size=bs, increment_block_size=bs)
            r = add(av, bv)
            np.testing.assert_array_equal(expected, r)

    # Multi-bit tree adders with configurable structure based on input bit width (2 configs tested here for the 9-bitwidth input)
    for c in [UnsignedHanCarlsonAdder, UnsignedKnowlesAdder, UnsignedLadnerFischerAdder]:
        for config in range(1, (math.ceil(math.log(N, 2))-2)+1):
            add = c(a, b, config_choice=config)
            r = add(av, bv)
            np.testing.assert_array_equal(expected, r)


def test_signed_add():
    """ Test signed adders """
    N = 9
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(-(2**(N-1)), 2**(N-1))
    bv = av.reshape(-1, 1)
    expected = av + bv

    # Non configurable multi-bit adders
    for c in [SignedPGRippleCarryAdder, SignedRippleCarryAdder, SignedConditionalSumAdder, SignedKoggeStoneAdder, SignedBrentKungAdder, SignedSklanskyAdder]:
        add = c(a, b)
        r = add(av, bv)
        np.testing.assert_array_equal(expected, r)

    # Multi-bit adders with configurable (uniform) logic blocks for parallel prefix computation (the adder will use the block_size argument it recognizes, others are ignored)
    for c in [SignedCarryLookaheadAdder, SignedCarrySkipAdder, SignedCarrySelectAdder, SignedCarryIncrementAdder]:
        for bs in range(1, N+1):
            add = c(a, b, cla_block_size=bs, bypass_block_size=bs, select_block_size=bs, increment_block_size=bs)
            r = add(av, bv)
            np.testing.assert_array_equal(expected, r)

    # Multi-bit tree adders with configurable structure based on input bit width (2 configs tested here for the 9-bitwidth input)
    for c in [SignedHanCarlsonAdder, SignedKnowlesAdder, SignedLadnerFischerAdder]:
        for config in range(1, (math.ceil(math.log(N, 2))-2)+1):
            add = c(a, b, config_choice=config)
            r = add(av, bv)
            np.testing.assert_array_equal(expected, r)



def test_unsigned_sub():
    """ Test unsigned subtractor """
    N = 9
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N)
    bv = av.reshape(-1, 1)
    expected = av - bv

    for c in [UnsignedRippleCarrySubtractor, UnsignedRippleBorrowSubtractor]:
        sub = c(a, b)
        r = sub(av, bv)
        np.testing.assert_array_equal(expected, r)


def test_signed_sub():
    """ Test signed subtractor """
    N = 9
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(-(2**(N-1)), 2**(N-1))
    bv = av.reshape(-1, 1)
    expected = av - bv

    # Non configurable multi-bit adders
    for c in [SignedRippleCarrySubtractor, SignedRippleBorrowSubtractor]:
        sub = c(a, b)
        r = sub(av, bv)
        np.testing.assert_array_equal(expected, r)

def test_mac():
    class MAC(GeneralCircuit):
        def __init__(self, a: Bus, b: Bus, r: Bus, prefix: str = "", name: str = "mac", **kwargs):
            super().__init__(prefix=prefix, name=name, out_N=2*a.N+1, inputs=[a, b, r], **kwargs)
            assert a.N == b.N
            assert r.N == 2 * a.N

            self.mul = self.add_component(UnsignedArrayMultiplier(a=Bus(wires_list=a.bus, prefix=a.prefix), b=Bus(wires_list=b.bus, prefix=b.prefix), prefix=self.prefix, name=f"u_arrmul{a.N}", inner_component=True))
            self.add = self.add_component(UnsignedRippleCarryAdder(a=Bus(wires_list=r.bus, prefix=r.prefix), b=self.mul.out, prefix=self.prefix, name=f"u_rca{r.N}", inner_component=True))
            self.out.connect_bus(connecting_bus=self.add.out)

    # usage
    mymac = MAC(Bus("a", 4), Bus("b", 4), Bus("acc", 8))

    av = np.arange(2**4)
    bv = np.arange(2**4).reshape(-1, 1)
    cv = np.arange(2**8).reshape(-1, 1, 1)

    r = mymac(av, bv, cv)
    expected = (av * bv) + cv
    np.testing.assert_array_equal(r, expected)


def test_direct():
    class err_circuit(GeneralCircuit):
        def __init__(self, a: Bus = Bus(), b: Bus = Bus(), prefix: str = "", name: str = "adder", inner_component: bool = False):
            super().__init__(prefix=prefix, name=name, out_N=(a.N + 1), inner_component=inner_component, inputs=[a, b])
            self.N = 1
            self.prefix = prefix
            self.a = Bus(prefix=a.prefix, wires_list=a.bus)
            self.b = Bus(prefix=b.prefix, wires_list=b.bus)
            self.out = Bus(self.prefix+"_out", self.N+1)

            a_0 = self.a[0]
            b_0 = self.b.get_wire(0)

            or_1 = OrGate(a_0, b_0, prefix=self.prefix+"_or"+str(self.get_instance_num(cls=OrGate)), parent_component=self)
            self.add_component(or_1)

            self.out.connect(0, a_0)
            self.out.connect(1, or_1.out)

    av = np.arange(0, 4).reshape(1, -1)
    bv = np.arange(0, 4).reshape(-1, 1)
    example = err_circuit(prefix="err_circuit", a=Bus("a", 2), b=Bus("b", 2))

    r = example(av, bv)
    expected = np.array([[0, 3, 0, 3], [2, 3, 2, 3], [0, 3, 0, 3], [2, 3, 2, 3]])
    np.testing.assert_equal(r, expected)
    print(r)


def test_wire_as_bus():
    """ accept a wire as a bus """
    class test_circuit(GeneralCircuit):
        def __init__(self, a: Wire, b: Wire, c: Bus, prefix="test_circuit", **kwargs):
            super().__init__(prefix=prefix, name="test_circuit", inputs=[a, b, c], out_N=1, **kwargs)
            g = self.add_component(AndGate(a, b, prefix="g2"))
            g2 = self.add_component(AndGate(g.out, c[0], prefix="g3"))
            g3 = self.add_component(AndGate(g2.out, c[1], prefix="g4"))
            self.out[0] = g3.out

    circ = test_circuit(Wire("a"), Wire("b"), Bus("c", 2), "c1")
    r = circ(np.array([0, 1]),
             np.array([0, 1]).reshape(-1, 1),
             np.arange(4).reshape(-1, 1, 1))
    assert r.sum() == 1
    assert r[-1, -1, -1] == 1

def test_unique():
    from ariths_gen.wire_components import Wire, Bus
    import pytest
    import sys
    from ariths_gen.one_bit_circuits.logic_gates import AndGate
    from ariths_gen.one_bit_circuits.one_bit_components import TwoOneMultiplexer
    from ariths_gen.core.arithmetic_circuits import GeneralCircuit

    class test_circuit(GeneralCircuit):
        def __init__(self, a: Bus, prefix="test_circuit", **kwargs):
            super().__init__(prefix=prefix, name="test_circuit", inputs=[a], out_N=1, **kwargs)
            g = self.add_component(AndGate(a[0], a[1], prefix="g2"))
            g2 = self.add_component(AndGate(g.out, a[2], prefix="g2"))
            g3 = self.add_component(AndGate(g2.out, g.out, prefix="g2"))
            self.out[0] = g3.out

    with pytest.raises(AssertionError):
        circ = test_circuit(Bus("a", 3), "c1")
        circ.get_v_code_flat(file_object=sys.stdout)

if __name__ == "__main__":
    test_unsigned_approxmul()
    test_unsigned_mul()
    test_signed_mul()
    test_unsigned_add()
    test_signed_add()
    test_unsigned_sub()
    test_signed_sub()
    test_mac()
    test_direct()
    test_wire_as_bus()
    print("Python tests were successful!")



from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    FullAdder,
    TwoOneMultiplexer
)
from ariths_gen.one_bit_circuits.logic_gates import (
    XorGate
)
import math


class UnsignedConditionalSumAdder(GeneralCircuit):
    """Class representing unsigned conditional sum adder.

    Conditional sum adder performs carry-select addition starting with
    groups of 1 bit and recursively doubling to N/2 bits.

    The individual input bit pairs are first added in a carry-select fashion (cin of 0 and 1) with block size of 1.
    Then in the subsequent two rows multiplexer pairs are used to properly select the sum and carry-out
    of the upper bit of each block of size 2 (higher bit of the previous 2-bits from the previous level).
    The next two rows follow the principle for blocks of size 4 and so on.

    The conditional sum adder shares the same structural characteristics as Sklansky tree adder.
    Because of using many multiplexers, the number of gates is significant when compared to other architectures.

    Another way on how to look at (and determine the number of) the multiplexers in the subsequent stages
    can be done by converting the input bit index (counting from 0) into its binary form. Each bit position with value 1
    signifies the stage position that contains multiplexers
    (which ones and how many of them depends on other input bit index characteristics deducible from the code below).

    ```
                B3A3            B2A2            B1A1            B0A0
                 ││              ││              ││              ││
              C4┌▼▼┐          C3┌▼▼┐          C2┌▼▼┐          C1┌▼▼┐
               ┌┤FA│◄─0      ┌──┤FA│◄─0        ┌┤FA│◄─0      ┌──┤FA│◄─Cin
               │└┬─┘         │  └┬─┘           │└┬─┘         │  └┬─┘
               │ │S3         │   │S2           │ │S1         │   │
               │ └───────┐   │   └───────┐     │ └───────┐   │   │
               └────────┐│   │           │     └────────┐│   │   │
                B3A3    ││   │  B2A2     │      B1A1    ││   │   │
                 ││     ││   │   ││      │       ││     ││   │   │
              C4┌▼▼┐    ││   │C3┌▼▼┐     │    C2┌▼▼┐    ││   │   │
               ┌┤FA│◄─1 ││   │┌─┤FA│◄─1  │     ┌┤FA│◄─1 ││   │   │
               │└┬─┘    ││   ││ └┬─┘     │     │└┬─┘    ││   │   │
               │ │S3    ││   ││  │S2     │     │ │S1    ││   │   │
              ┌┼─┴┐ ┌───┤│   ││  │       │     │ └┐ ┌───┘│   │   │
              │├─┐│ │┌──┼┤   ││  │   ┌───┘     └─┐│ │┌───┘   │   │
              ││ ││ ││  ││   ││  │   │           ││ ││       │   │
              ││┌▼▼─▼▼┐ ││   ││  │   │          ┌▼▼─▼▼┐      │   │
              │││ 2:1 │ ││SEL││  │   │          │ 2:1 │   SEL│   │
    Stage 0:  │││MUXes│◄┼┼───┘│  │   │          │MUXes│◄─────┘   │
              │││ CS  │ ││    │  │   │          │ CS  │          │
              ││└┬───┬┘ ││    │  │   │          └┬───┬┘          │
              ││ │C4 │S3││    │  │   │           │C2 │           │
              ││ └───┼──┼┼────┼┐ │   │           │   │           │
              ││     └──┼┼────┼┼┐│   │           │   │           │
              └┼──┐ ┌───┘│    ││││   │           │   │           │
               └─┐│ │┌───┘    ││││   │           │   │           │
                 ││ ││        ││││   │           │   │           │
                ┌▼▼─▼▼┐       ││││   │           │   │           │
                │ 2:1 │   SEL ││││   │           │   │           │
    Stage 0:    │MUXes│◄──────┘│││   │           │   │           │
                │ CS  │        │││   │           │   │           │
                └┬───┬┘        │││   │           │   │           │
                 │C4 │S3       │││   │           │   │           │
                 │┌──┘         │││   │           │   │           │
                 ││ ┌──────────┘││   │           │   │           │
                 ││ │┌──────────┘│   │           │   │           │
                 ││ ││       ┌───┼───┼───────────┤   │           │
                ┌▼▼─▼▼┐      │  ┌▼───▼┐          │   │           │
                │ 2:1 │   SEL│  │ 2:1 │   SEL    │   │           │
    Stage 1:    │MUXes│◄─────┘  │MUXes│◄─────────┘   │           │
                │ CS  │         │ S   │              │           │
                └┬───┬┘         └────┬┘              │           │
                 │C4 │               │               │           │
                 │   │               │               │           │
                 ▼   ▼               ▼               ▼           ▼
                Cout S3              S2              S1          S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned cosa. Defaults to "".
        name (str, optional): Name of unsigned cosa. Defaults to "u_cosa".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_cosa", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)
        cin = ConstantWireValue0()

        # Lists of lists containing all sum/carry wires
        self.sum_sig = [([], []) for i in range(self.N)]
        self.carry_sig = [([], []) for i in range(self.N+1)]  # +1 because of cin
        self.carry_sig[0][0].append(cin)

        for i_wire in range(self.N):
            if i_wire == 0:
                self.add_component(FullAdder(a=self.a.get_wire(i_wire), b=self.b.get_wire(i_wire), c=cin, prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder))))
                self.sum_sig[i_wire][0].append(self.get_previous_component().get_sum_wire())
                self.carry_sig[i_wire+1][0].append(self.get_previous_component().get_carry_wire())

                self.out.connect(i_wire, self.get_previous_component().get_sum_wire())
            else:
                # Create FA with cin set to ConstantValue0 (and ConstantValue1 respectively) and save the generated sum and carry wires to the appropriate positions within the sum_sig and carry_sig lists to allow easier access when doing interconnections between muxes
                self.add_component(FullAdder(a=self.a.get_wire(i_wire), b=self.b.get_wire(i_wire), c=ConstantWireValue0(), prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder))))
                self.sum_sig[i_wire][0].append(self.get_previous_component().get_sum_wire())
                self.carry_sig[i_wire+1][0].append(self.get_previous_component().get_carry_wire())
                self.add_component(FullAdder(a=self.a.get_wire(i_wire), b=self.b.get_wire(i_wire), c=ConstantWireValue1(), prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder))))
                self.sum_sig[i_wire][1].append(self.get_previous_component().get_sum_wire())
                self.carry_sig[i_wire+1][1].append(self.get_previous_component().get_carry_wire())

                # Determine the number of 2:1 muxes stages to choose which of the fa sum and carry bit to propagate further
                binary_form = bin(i_wire)[2:][::-1]
                index_stages = len(binary_form)
                prev_stage_int_value = 0
                # Create approprite structure of muxes and interconnect wires according to the bit index and stage
                for stage, value in enumerate(binary_form):
                    if value == '1':
                        # Both sum and carry muxes in stages
                        if i_wire % 2 != 0 or i_wire == self.N-1:  # Odd-numbered bit indexes
                            if stage == index_stages-1 and (math.log(i_wire+1, 2).is_integer() or i_wire == self.N-1):  # Stage with only one pair of sum and carry multiplexers
                                # Sum
                                self.add_component(TwoOneMultiplexer(a=self.sum_sig[i_wire][0][-1], b=self.sum_sig[i_wire][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][0][-1], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))
                                self.sum_sig[i_wire][0].append(self.get_previous_component().out.get_wire())
                                # Carry
                                self.add_component(TwoOneMultiplexer(a=self.carry_sig[i_wire+1][0][-1], b=self.carry_sig[i_wire+1][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][0][-1], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))
                                self.carry_sig[i_wire+1][0].append(self.get_previous_component().out.get_wire())

                                self.out.connect(i_wire, self.sum_sig[i_wire][0][-1])
                            elif stage == index_stages-1:  # Stage with only one sum multiplexer
                                self.add_component(TwoOneMultiplexer(a=self.sum_sig[i_wire][0][-1], b=self.sum_sig[i_wire][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][0][-1], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))
                                self.sum_sig[i_wire][0].append(self.get_previous_component().out.get_wire())

                                self.out.connect(i_wire, self.sum_sig[i_wire][0][-1])
                            else:  # Stage with two pairs of sum and carry multiplexers
                                # Sum
                                self.add_component(TwoOneMultiplexer(a=self.sum_sig[i_wire][0][-1], b=self.sum_sig[i_wire][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][0][-1], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))
                                self.add_component(TwoOneMultiplexer(a=self.sum_sig[i_wire][0][-1], b=self.sum_sig[i_wire][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][1][-1], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))

                                self.sum_sig[i_wire][0].append(self.get_previous_component(2).out.get_wire())
                                self.sum_sig[i_wire][1].append(self.get_previous_component().out.get_wire())
                                # Carry
                                self.add_component(TwoOneMultiplexer(a=self.carry_sig[i_wire+1][0][-1], b=self.carry_sig[i_wire+1][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][0][-1], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))
                                self.add_component(TwoOneMultiplexer(a=self.carry_sig[i_wire+1][0][-1], b=self.carry_sig[i_wire+1][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][1][-1], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))

                                self.carry_sig[i_wire+1][0].append(self.get_previous_component(2).out.get_wire())
                                self.carry_sig[i_wire+1][1].append(self.get_previous_component().out.get_wire())

                        # Only sum muxes in all stages
                        else:  # Even-numbered bit indexes
                            if stage == index_stages-1:  # Stage with only one sum multiplexer
                                self.add_component(TwoOneMultiplexer(a=self.sum_sig[i_wire][0][-1], b=self.sum_sig[i_wire][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][0][-1], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))
                                self.sum_sig[i_wire][0].append(self.get_previous_component().out.get_wire())
                                self.out.connect(i_wire, self.sum_sig[i_wire][0][-1])
                            else:  # Stage with two sum multiplexers
                                self.add_component(TwoOneMultiplexer(a=self.sum_sig[i_wire][0][-1], b=self.sum_sig[i_wire][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][0][stage], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))
                                self.sum_sig[i_wire][0].append(self.get_previous_component().out.get_wire())

                                self.add_component(TwoOneMultiplexer(a=self.sum_sig[i_wire][0][-1], b=self.sum_sig[i_wire][1][-1], c=self.carry_sig[i_wire-prev_stage_int_value][1][stage], prefix=self.prefix+"_mux2to1_"+str(self.get_instance_num(cls=TwoOneMultiplexer))))
                                self.sum_sig[i_wire][1].append(self.get_previous_component().out.get_wire())

                        prev_stage_int_value += 2**stage
            if i_wire == self.N-1:
                self.out.connect(self.N, self.carry_sig[i_wire+1][0][-1])


class SignedConditionalSumAdder(UnsignedConditionalSumAdder, GeneralCircuit):
    """Class representing signed conditional sum adder.

    Conditional sum adder performs carry-select addition starting with
    groups of 1 bit and recursively doubling to N/2 bits.

    The individual input bit pairs are first added in a carry-select fashion (cin of 0 and 1) with block size of 1.
    Then in the subsequent two rows multiplexer pairs are used to properly select the sum and carry-out
    of the upper bit of each block of size 2 (higher bit of the previous 2-bits from the previous level).
    The next two rows follow the principle for blocks of size 4 and so on.
    At last XOR gates are used to ensure proper sign extension.

    The conditional sum adder shares the same structural characteristics as Sklansky tree adder.
    Because of using many multiplexers, the number of gates is significant when compared to other architectures.

    Another way on how to look at (and determine the number of) the multiplexers in the subsequent stages
    can be done by converting the input bit index (counting from 0) into its binary form. Each bit position with value 1
    signifies the stage position that contains multiplexers
    (which ones and how many of them depends on other input bit index characteristics deducible from the code below).

        ```
                B3A3            B2A2            B1A1            B0A0
                 ││              ││              ││              ││
              C4┌▼▼┐          C3┌▼▼┐          C2┌▼▼┐          C1┌▼▼┐
               ┌┤FA│◄─0      ┌──┤FA│◄─0        ┌┤FA│◄─0      ┌──┤FA│◄─Cin
               │└┬─┘         │  └┬─┘           │└┬─┘         │  └┬─┘
               │ │S3         │   │S2           │ │S1         │   │
               │ └───────┐   │   └───────┐     │ └───────┐   │   │
               └────────┐│   │           │     └────────┐│   │   │
                B3A3    ││   │  B2A2     │      B1A1    ││   │   │
                 ││     ││   │   ││      │       ││     ││   │   │
              C4┌▼▼┐    ││   │C3┌▼▼┐     │    C2┌▼▼┐    ││   │   │
               ┌┤FA│◄─1 ││   │┌─┤FA│◄─1  │     ┌┤FA│◄─1 ││   │   │
               │└┬─┘    ││   ││ └┬─┘     │     │└┬─┘    ││   │   │
               │ │S3    ││   ││  │S2     │     │ │S1    ││   │   │
              ┌┼─┴┐ ┌───┤│   ││  │       │     │ └┐ ┌───┘│   │   │
              │├─┐│ │┌──┼┤   ││  │   ┌───┘     └─┐│ │┌───┘   │   │
              ││ ││ ││  ││   ││  │   │           ││ ││       │   │
              ││┌▼▼─▼▼┐ ││   ││  │   │          ┌▼▼─▼▼┐      │   │
              │││ 2:1 │ ││SEL││  │   │          │ 2:1 │   SEL│   │
    Stage 0:  │││MUXes│◄┼┼───┘│  │   │          │MUXes│◄─────┘   │
              │││ CS  │ ││    │  │   │          │ CS  │          │
              ││└┬───┬┘ ││    │  │   │          └┬───┬┘          │
              ││ │C4 │S3││    │  │   │           │C2 │           │
              ││ └───┼──┼┼────┼┐ │   │           │   │           │
              ││     └──┼┼────┼┼┐│   │           │   │           │
              └┼──┐ ┌───┘│    ││││   │           │   │           │
               └─┐│ │┌───┘    ││││   │           │   │           │
                 ││ ││        ││││   │           │   │           │
                ┌▼▼─▼▼┐       ││││   │           │   │           │
                │ 2:1 │   SEL ││││   │           │   │           │
    Stage 0:    │MUXes│◄──────┘│││   │           │   │           │
                │ CS  │        │││   │           │   │           │
                └┬───┬┘        │││   │           │   │           │
                 │C4 │S3       │││   │           │   │           │
                 │┌──┘         │││   │           │   │           │
                 ││ ┌──────────┘││   │           │   │           │
                 ││ │┌──────────┘│   │           │   │           │
                 ││ ││       ┌───┼───┼───────────┤   │           │
                ┌▼▼─▼▼┐      │  ┌▼───▼┐          │   │           │
                │ 2:1 │   SEL│  │ 2:1 │   SEL    │   │           │
    Stage 1:    │MUXes│◄─────┘  │MUXes│◄─────────┘   │           │
                │ CS  │         │ S   │              │           │
                └┬───┬┘         └────┬┘              │           │
                 │C4 │               │               │           │
             ┌───▼──┐│               │               │           │
             │ SIGN ││               │               │           │
             │Extend││               │               │           │
             └───┬──┘│               │               │           │
                 │   │               │               │           │
                 ▼   ▼               ▼               ▼           ▼
                Cout S3              S2              S1          S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed cosa. Defaults to "".
        name (str, optional): Name of signed cosa. Defaults to "s_cosa".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_cosa", **kwargs):
        # Signed bus length extension
        N = max(a.N, b.N)
        a.bus_extend(N=N, prefix=a.prefix, desired_extension_wire=a.get_wire(a.N-1))
        b.bus_extend(N=N, prefix=b.prefix, desired_extension_wire=b.get_wire(b.N-1))
        
        super().__init__(a=a, b=b, prefix=prefix, name=name, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        self.add_component(XorGate(self.a.get_wire(self.N-1), self.b.get_wire(self.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate, count_disabled_gates=False)), parent_component=self))
        self.add_component(XorGate(self.get_previous_component().out, self.out.get_wire(-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate, count_disabled_gates=False)), parent_component=self))
        self.out.connect(self.N, self.get_previous_component().out)
